# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import csv
import copy
import math
import six

# Morphology nodes have the following fields. SWC fields are numeric.
NODE_ID = 'id'
NODE_TYPE = 'type'
NODE_X = 'x'
NODE_Y = 'y'
NODE_Z = 'z'
NODE_R = 'radius'
NODE_PN = 'parent'
SWC_COLUMNS = [NODE_ID, NODE_TYPE, NODE_X, NODE_Y, NODE_Z, NODE_R, NODE_PN]

NODE_TREE_ID = 'tree_id'
NODE_CHILDREN = 'children'

# shorthand for dictionary entries, to shorten sometimes long code lines
_N = NODE_ID
_TYP = NODE_TYPE
_X = NODE_X
_Y = NODE_Y
_Z = NODE_Z
_R = NODE_R
_P = NODE_PN
_C = NODE_CHILDREN
_TID = NODE_TREE_ID


########################################################################
def read_swc(file_name, columns="NOT_USED", numeric_columns="NOT_USED"):
    """
    Read in an SWC file and return a Morphology object.

    Parameters
    ----------
    file_name: string
        SWC file name.

    Returns
    -------
    Morphology
        A Morphology instance.
    """
    compartments = []
    line_num = 1
    try:
        with open(file_name, "r") as f:
            for line in f:
                # remove comments
                if line.lstrip().startswith('#'):
                    continue
                # read values. expected SWC format is:
                #   ID, type, x, y, z, rad, parent
                # x, y, z and rad are floats. the others are ints
                toks = line.split(' ')
                vals = Compartment({
                    NODE_ID: int(toks[0]),
                    NODE_TYPE: int(toks[1]),
                    NODE_X: float(toks[2]),
                    NODE_Y: float(toks[3]),
                    NODE_Z: float(toks[4]),
                    NODE_R: float(toks[5]),
                    NODE_PN: int(toks[6].rstrip())
                })
                # store this compartment
                compartments.append(vals)
                # increment line number (used for error reporting only)
                line_num += 1
    except ValueError:
        err = "File not recognized as valid SWC file.\n"
        err += "Problem parsing line %d\n" % line_num
        if line is not None:
            err += "Content: '%s'\n" % line
        raise IOError(err)

    return Morphology(compartment_list=compartments)


########################################################################
########################################################################
class Compartment(dict):
    """
    A dictionary class storing information about a single morphology node
    """

    def __init__(self, *args, **kwargs):
        super(Compartment, self).__init__(*args, **kwargs)
        if (NODE_ID not in self or
                NODE_TYPE not in self or
                NODE_X not in self or
                NODE_Y not in self or
                NODE_Z not in self or
                NODE_R not in self or
                NODE_PN not in self):
            raise ValueError(
                "Compartment was not initialized with requisite fields")
        # Each unconnected graph has its own ID. This is the ID
        #   of graph that the node resides in
        self[NODE_TREE_ID] = -1

        # IDs of child nodes
        self[NODE_CHILDREN] = []

    def print_node(self):
        """ print out compartment information with field names """
        print("%d %d %.4f %.4f %.4f %.4f %d %s %d" % (self[_N], self[_TYP], self[
              _X], self[_Y], self[_Z], self[_R], self[_P], str(self[_C]), self[_TID]))


class Morphology(object):
    """
    Keep track of the list of compartments in a morphology and provide
    a few helper methods (soma, tree information, pruning, etc).
    """

    SOMA = 1
    AXON = 2
    DENDRITE = 3
    BASAL_DENDRITE = 3
    APICAL_DENDRITE = 4

    NODE_TYPES = [SOMA, AXON, DENDRITE, BASAL_DENDRITE, APICAL_DENDRITE]

    def __init__(self, compartment_list=None, compartment_index=None):
        """
        Try to initialize from a list of compartments first, then from
        a dictionary indexed by compartment id if that fails, and finally just
        leave everything empty.

        Parameters
        ----------
        compartment_list: list
            list of compartment dictionaries

        compartment_index: dict
            dictionary of compartments indexed by id
        """
        self._compartment_list = []
        self._compartment_index = {}

        ##############################################
        # define tree list here for clarity, even though it's reset below
        #   when nodes are assigned
        self._tree_list = []

        ##############################################
        # construct the compartment list and index
        # first try to do so using the compartment list, then try using
        #   the compartment index and if that fails then complain
        if compartment_list:
            self.compartment_list = compartment_list
        elif compartment_index:
            self.compartment_index = compartment_index
        ##############################################
        # verify morphology is consistent with morphology rules (e.g.,
        #   no dendrite branching from an axon)
        num_errors = self._check_consistency()
        if num_errors > 0:
            raise ValueError("Morphology appears to be inconsistent")
        ##############################################
        # root node (this must be part of the soma)
        self._soma = None
        for i in range(len(self.compartment_list)):
            seg = self.compartment_list[i]
            if seg[NODE_TYPE] == Morphology.SOMA and seg[NODE_PN] < 0:
                if self._soma is not None:
                    raise ValueError("Multiple somas detected in SWC file")
                self._soma = seg

    ####################################################################
    ####################################################################
    # class properties, and helper functions for them

    @property
    def compartment_list(self):
        """ Return the compartment list.  This is a property to ensure that the
        compartment list and compartment index are in sync. """
        return self._compartment_list

    @compartment_list.setter
    def compartment_list(self, compartment_list):
        """ Update the compartment list.  Update the compartment index. """
        self._set_compartments(compartment_list)

    @property
    def compartment_index(self):
        """ Return the compartment index.  This is a property to ensure that the
        compartment list and compartment index are in sync. """
        return self._compartment_index

    @compartment_index.setter
    def compartment_index(self, compartment_index):
        """ Update the compartment index.  Update the compartment list. """
        self._set_compartments(compartment_index.values())

    @property
    def num_trees(self):
        """ Return the number of trees in the morphology. A tree is
        defined as everything following from a single root compartment. """
        return len(self._tree_list)

    # TODO add filter for number of nodes of a particular type
    @property
    def num_nodes(self):
        """ Return the number of compartments in the morphology. """
        return len(self.compartment_list)

    # internal function
    def _set_compartments(self, compartment_list):
        """
        take a list of SWC-like objects and turn those into morphology
        nodes need to be able to initialize from a list supplied by an SWC
        file while also being able to initialize from the compartment list
        of an existing Morphology object. As nodes in a morphology object
        contain reference to nodes in that object, make a shallow copy
        of input nodes and overwrite known references (ie, the
        'children' array)
        """
        self._compartment_list = []
        for obj in compartment_list:
            seg = copy.copy(obj)
            seg[NODE_TREE_ID] = -1
            seg[NODE_CHILDREN] = []
            self._compartment_list.append(seg)
        # list data now set. remove holes in sequence and re-index
        self._reconstruct()

    @property
    def soma(self):
        """ Returns root node of soma, if present"""
        return self._soma

    @property
    def root(self):
        """ [deprecated] Returns root node of soma, if present. Use 'soma' instead of 'root'"""
        return self._soma

    ####################################################################
    ####################################################################
    # tree and node access

    def tree(self, n):
        """
        Returns a list of all Morphology Nodes within the specified
        tree. A tree is defined as a fully connected graph of nodes.
        Each tree has exactly one root.

        Parameters
        ----------
        n: integer
            ID of desired tree

        Returns
        -------
        A list of all morphology objects in the specified tree, or None
        if the tree doesn't exist
        """
        if n < 0 or n >= len(self._tree_list):
            return None
        return self._tree_list[n]

    def node(self, n):
        """
        Returns the morphology node having the specified ID.

        Parameters
        ----------
        n: integer
            ID of desired node

        Returns
        -------
        A morphology object having the specified ID, or None if such a
        node doesn't exist
        """
        # undocumented feature -- if a node is supplied instead of a
        #   node ID, the node is returned and no error is triggered
        return self._resolve_node_type(n)

    def parent_of(self, seg):
        """ Returns parent of the specified node.

        Parameters
        ----------
        seg: integer or Morphology Object
            The ID of the child node, or the child node itself

        Returns
        -------
        A morphology object, or None if no parent exists or if the
        specified node ID doesn't exist
        """
        # if ID passed in, make sure it's converted to a compartment
        # don't trap for exception here -- if supplied segment is
        #   incorrect, make sure the user knows about it
        seg = self._resolve_node_type(seg)
        # return parent of specified node
        if seg is not None and seg[NODE_PN] >= 0:
            return self._compartment_list[seg[NODE_PN]]
        return None

    def children_of(self, seg):
        """ Returns a list of the children of the specified node

        Parameters
        ----------
        seg: integer or Morphology Object
            The ID of the parent node, or the parent node itself

        Returns
        -------
        A list of the child morphology objects. If the ID of the parent
        node is invalid, None is returned.
        """
        seg = self._resolve_node_type(seg)
        return [self._compartment_list[c] for c in seg[NODE_CHILDREN]]

    ###################################################################
    ###################################################################
    # Information querying and data manipulation

    # internal function. takes an integer and returns the node having
    #   that ID. IF a node is passed in instead, it is returned
    def _resolve_node_type(self, seg):
        # if compartment passed then we don't need to convert anything
        # if compartment not passed, try converting value to int
        #   and using that as an index
        if not isinstance(seg, Compartment):
            try:
                seg = int(seg)
                if seg < 0 or seg >= len(self._compartment_list):
                    return None
                seg = self._compartment_list[seg]
            except ValueError:
                raise TypeError(
                    "Object not recognized as morphology node or index")
        return seg

    def change_parent(self, child, parent):
        """ Change the parent of a node. The child node is adjusted to
        point to the new parent, the child is taken off of the previous
        parent's child list, and it is added to the new parent's child list.

        Parameters
        ----------
        child: integer or Morphology Object
            The ID of the child node, or the child node itself

        parent: integer or Morphology Object
            The ID of the parent node, or the parent node itself

        Returns
        -------
        Nothing
        """
        child_seg = self._resolve_node_type(child)
        parent_seg = self._resolve_node_type(parent)
        # if child has former parent, remove it from parent's child list
        if child_seg[NODE_PN] >= 0:
            old_par = self.node(child_seg[NODE_PN])
            old_par[NODE_CHILDREN].remove(child_seg[NODE_ID])
        parent_seg[NODE_CHILDREN].append(child_seg[NODE_ID])
        child_seg[NODE_PN] = parent_seg[NODE_ID]

    # returns a list of nodes located within dist of x,y,z
    def find(self, x, y, z, dist, node_type=None):
        """ Returns a list of Morphology Objects located within 'dist'
        of coordinate (x,y,z). If node_type is specified, the search
        will be constrained to return only nodes of that type.

        Parameters
        ----------
        x, y, z: float
            The x,y,z coordinates from which to search around

        dist: float
            The search radius

        node_type: enum (optional)
            One of the following constants: SOMA, AXON, DENDRITE,
            BASAL_DENDRITE or APICAL_DENDRITE

        Returns
        -------
        A list of all Morphology Objects matching the search criteria
        """
        found = []
        for seg in self.compartment_list:
            dx = seg[NODE_X] - x
            dy = seg[NODE_Y] - y
            dz = seg[NODE_Z] - z
            if math.sqrt(dx * dx + dy * dy + dz * dz) <= dist:
                if node_type is None or seg[NODE_TYPE] == node_type:
                    found.append(seg)
        return found

    def compartment_list_by_type(self, compartment_type):
        """ Return an list of all compartments having the specified
        compartment type.

        Parameters
        ----------
        compartment_type: int
            Desired compartment type

        Returns
        -------
        A list of of Morphology Objects
        """
        return [x for x in self._compartment_list if x[NODE_TYPE] == compartment_type]

    def compartment_index_by_type(self, compartment_type):
        """ Return an dictionary of compartments indexed by id that all have
        a particular compartment type.

        Parameters
        ----------
        compartment_type: int
            Desired compartment type

        Returns
        -------
        A dictionary of Morphology Objects, indexed by ID
        """
        return {c[NODE_ID]: c for c in self._compartment_list if c[NODE_TYPE] == compartment_type}

    def save(self, file_name):
        """ Write this morphology out to an SWC file

        Parameters
        ----------
        file_name: string
            desired name of your SWC file
        """
        f = open(file_name, "w")
        f.write("#n,type,x,y,z,radius,parent\n")
        for seg in self.compartment_list:
            f.write("%d %d " % (seg[NODE_ID], seg[NODE_TYPE]))
            f.write("%0.4f " % seg[NODE_X])
            f.write("%0.4f " % seg[NODE_Y])
            f.write("%0.4f " % seg[NODE_Z])
            f.write("%0.4f " % seg[NODE_R])
            f.write("%d\n" % seg[NODE_PN])
        f.close()

    # keep for backward compatibility, but don't publish in docs
    def write(self, file_name):
        self.save(file_name)

    def sparsify(self, modulo, compress_ids=False):
        """ Return a new Morphology object that has a given number of non-leaf,
        non-root nodes removed.  IDs can be reassigned so as to be continuous.

        Parameters
        ----------
        modulo: int
           keep 1 out of every modulo nodes.

        compress_ids: boolean
           Reassign ids so that ids are continuous (no missing id numbers).

        Returns
        -------
        Morphology
            A new morphology instance
        """
        compartments = self.compartment_index
        root = self.root
        keep = {}
        # figure out which compartments to toss
        ct = 0
        for i, c in six.iteritems(compartments):
            pid = c[NODE_PN]
            cid = c[NODE_ID]
            ctype = c[NODE_TYPE]
            # keep the root, soma, junctions, and the first child of the root
            # (for visualization)
            if pid < 0 or len(c[NODE_CHILDREN]) != 1 or pid == root[NODE_ID] or ctype == Morphology.SOMA:
                keep[cid] = True
            else:
                keep[cid] = (ct % modulo) == 0
            ct += 1

        # hook children up to their new parents
        for i, c in six.iteritems(compartments):
            comp_id = c[NODE_ID]
            if keep[comp_id] is False:
                parent_id = c[NODE_PN]
                while keep[parent_id] is False:
                    parent_id = compartments[parent_id][NODE_PN]
                for child_id in c[NODE_CHILDREN]:
                    compartments[child_id][NODE_PN] = parent_id

        # filter out the orphans
        sparsified_compartments = {k: v for k,
                                   v in six.iteritems(compartments) if keep[k]}
        if compress_ids:
            ids = sorted(sparsified_compartments.keys(), key=lambda x: int(x))
            id_hash = {fid: str(i + 1) for i, fid in enumerate(ids)}
            id_hash[-1] = -1
            # build the final compartment index
            out_compartments = {}
            for cid, compartment in six.iteritems(sparsified_compartments):
                compartment[NODE_ID] = id_hash[cid]
                compartment[NODE_PN] = id_hash[compartment[NODE_PN]]
                out_compartments[compartment[NODE_ID]] = compartment
            return Morphology(compartment_index=out_compartments)
        else:
            return Morphology(compartment_index=sparsified_compartments)

    ####################################################################
    ####################################################################
    def _reconstruct(self):
        """
        internal function that restructures data and establishes
        appropriate internal linking. data is re-order, removing 'holes'
        in sequence so that each object ID corresponds to its position
        in compartment list. trees are (re)calculated
        parent-child indices are recalculated as is compartment table
        construct a map between new and old IDs
        """
        remap = {}
        # everything defaults to root. this way if a parent was deleted
        #   the child will become a new root
        for i in range(len(self.compartment_list)):
            remap[i] = -1
        # map old old node numbers to new ones. reset n to the new ID
        #   and put node in new list
        new_id = 0
        tmp_list = []
        for seg in self.compartment_list:
            if seg is not None:
                remap[seg[NODE_ID]] = new_id
                seg[NODE_ID] = new_id
                tmp_list.append(seg)
                new_id += 1
        # use map to reset parent values. copy objs to new list
        for seg in tmp_list:
            if seg[NODE_PN] >= 0:
                seg[NODE_PN] = remap[seg[NODE_PN]]
        # replace compartment list with newly created node list
        self._compartment_list = tmp_list
        # reconstruct parent/child relationship links
        # forget old relations
        for seg in self.compartment_list:
            seg[NODE_CHILDREN] = []
        # add each object to its parents child list
        for seg in self.compartment_list:
            par_num = seg[NODE_PN]
            if par_num >= 0:
                self.compartment_list[par_num][
                    NODE_CHILDREN].append(seg[NODE_ID])
        # update tree lists
        self._separate_trees()
        ############################
        # Rebuild internal index and links between parents and children
        self._compartment_index = {
            c[NODE_ID]: c for c in self.compartment_list}
        # compartment list is complete and sequential so don't need index
        #   to resolve relationships
        # for each node, reset children array
        # for each node, add self to parent's child list
        for seg in self._compartment_list:
            seg[NODE_CHILDREN] = []
        for seg in self._compartment_list:
            if seg[NODE_PN] >= 0:
                self._compartment_list[seg[NODE_PN]][
                    NODE_CHILDREN].append(seg[NODE_ID])
        # verify that each node ID is the same as its position in the
        #   compartment list
        for i in range(len(self.compartment_list)):
            if i != self.node(i)[NODE_ID]:
                raise RuntimeError(
                    "Internal error detected -- compartment list not properly formed")

    def append(self, node_list):
        """ Add additional nodes to this Morphology. Those nodes must
        originate from another morphology object.

        Parameters
        ----------
        node_list: list of Morphology nodes
        """
        # construct a map between new and old IDs of added nodes
        remap = {}
        for i in range(len(node_list)):
            remap[i] = -1
        # map old old node numbers to new ones. reset n to the new ID
        # append new nodes to existing node list
        old_count = len(self.compartment_list)
        new_id = old_count
        for seg in node_list:
            if seg is not None:
                remap[seg[NODE_ID]] = new_id
                seg[NODE_ID] = new_id
                self._compartment_list.append(seg)
                new_id += 1
        # use map to reset parent values. copy objs to new list
        for i in range(old_count, len(self.compartment_list)):
            seg = self.compartment_list[i]
            if seg[NODE_PN] >= 0:
                seg[NODE_PN] = remap[seg[NODE_PN]]
        self._reconstruct()

    def stumpify_axon(self, count=10):
        """ Remove all axon compartments except the first 'count'
        nodes, as counted from the connected axon root.

        Parameters
        ----------
        count: Integer
            The length of the axon 'stump', in number of compartments
        """
        # find connected axon root
        axon_root = None
        for seg in self.compartment_list:
            if seg[NODE_TYPE] == Morphology.AXON:
                par_id = seg[NODE_PN]
                if par_id >= 0:
                    par = self.compartment_list[par_id]
                    if par[NODE_TYPE] != Morphology.AXON:
                        axon_root = seg
                        break
        if axon_root is None:
            return
        # flag the first 'count' nodes from the axon root
        ax = axon_root
        for i in range(count):
            # ignore bifurcations -- go 'count' deep on one line only
            ax["flag"] = i
            children = ax[NODE_CHILDREN]
            if len(children) > 0:
                ax = children[0]
        # strip out all axons that aren't flagged
        for i in range(len(self.compartment_list)):
            seg = self.compartment_list[i]
            if seg[NODE_TYPE] == Morphology.AXON:
                if "flag" not in seg:
                    self.compartment_list[i] = None
        self._reconstruct()

    # strip out everything but the soma and the specified SWC type
    def strip_all_other_types(self, node_type, keep_soma=True):
        """ Strips everything from the morphology except for the
        specified type.
        Parent and child relationships are updated accordingly, creating
        new roots when necessary.

        Parameters
        ----------
        node_type: enum
            The compartment type to keep in the morphology.
            Use one of the following constants: SOMA, AXON, DENDRITE,
            BASAL_DENDRITE, or APICAL_DENDRITE

        keep_soma: Boolean (optional)
            True (default) if soma nodes should remain in the
            morpyhology, and False if the soma should also be stripped
        """
        flagged_for_removal = {}
        # scan nodes and see which ones should be removed. keep a record
        #   of them
        for seg in self.compartment_list:
            if seg[NODE_TYPE] == node_type:
                remove = False
            elif seg[NODE_TYPE] == 1 and keep_soma:
                remove = False
            else:
                remove = True
            if remove:
                flagged_for_removal[seg[NODE_ID]] = True
        # remove selected nodes andreset parent links
        for i in range(len(self.compartment_list)):
            seg = self.compartment_list[i]
            if seg[NODE_ID] in flagged_for_removal:
                # eliminate node
                self.compartment_list[i] = None
            elif seg[NODE_PN] in flagged_for_removal:
                # parent was eliminated. make this a new root
                seg[NODE_PN] = -1
        self._reconstruct()

    # strip out the specified SWC type
    def strip_type(self, node_type):
        """ Strips all compartments of the specified type from the
        morphology.
        Parent and child relationships are updated accordingly, creating
        new roots when necessary.

        Parameters
        ----------
        node_type: enum
            The compartment type to strip from the morphology.
            Use one of the following constants: SOMA, AXON, DENDRITE,
            BASAL_DENDRITE, or APICAL_DENDRITE
        """
        flagged_for_removal = {}
        for seg in self.compartment_list:
            if seg[NODE_TYPE] == node_type:
                remove = True
            else:
                remove = False
            if remove:
                flagged_for_removal[seg[NODE_ID]] = True
        for i in range(len(self.compartment_list)):
            seg = self.compartment_list[i]
            if seg[NODE_ID] in flagged_for_removal:
                # eliminate node
                self.compartment_list[i] = None
            elif seg[NODE_PN] in flagged_for_removal:
                # parent was eliminated. make this a new root
                seg[NODE_PN] = -1
        self._reconstruct()

    # strip out the specified SWC type
    def convert_type(self, old_type, new_type):
        """ Converts all compartments from one type to another.
        Nodes of the original type are not affected so this
        procedure can also be used as a merge procedure.

        Parameters
        ----------
        old_type: enum
            The compartment type to be changed.
            Use one of the following constants: SOMA, AXON, DENDRITE,
            BASAL_DENDRITE, or APICAL_DENDRITE

        new_type: enum
            The target compartment type.
            Use one of the following constants: SOMA, AXON, DENDRITE,
            BASAL_DENDRITE, or APICAL_DENDRITE
        """
        for seg in self.compartment_list:
            if seg[NODE_TYPE] == old_type:
                seg[NODE_TYPE] = new_type

    def apply_affine(self, aff, scale=None):
        """ Apply an affine transform to all compartments in this
        morphology. Node radius is adjusted as well.

        Format of the affine matrix is:

        [x0 y0 z0]  [tx]
        [x1 y1 z1]  [ty]
        [x2 y2 z2]  [tz]

        where the left 3x3 the matrix defines the affine rotation
        and scaling, and the right column is the translation
        vector.

        The matrix must be collapsed and stored in a list as follows:

        [x0 y0, z0, x1, y1, z1, x2, y2, z2, tx, ty, tz]

        Parameters
        ----------
        aff: 3x4 array of floats (python 2D list, or numpy 2D array)
            the transformation matrix
        """
        # In addition to transforming the locations of the morphology
        #   nodes, the radius of each node must be adjusted.
        # There are 2 ways to measure scale from a transform. Assuming
        #   an isotropic transform, the scale is the cube root of the
        #   matrix determinant. The other ways is to measure scale
        #   independently along each axis.
        # For now, the node radius is only updated based on the average
        #   scale along all 3 axes (eg, isotropic assumption), so calculate
        #   scale using the determinant
        #
        if scale is None:
            # calculate the determinant
            det0 = aff[0] * (aff[4] * aff[8] - aff[5] * aff[7])
            det1 = aff[1] * (aff[3] * aff[8] - aff[5] * aff[6])
            det2 = aff[2] * (aff[3] * aff[7] - aff[4] * aff[6])
            det = det0 + det1 + det2
            # determinant is change of volume that occurred during transform
            # assume equal scaling along all axes. take 3rd root to get
            #   scale factor
            det_scale = math.pow(abs(det), 1.0 / 3.0)
            # measure scale along each axis
            # keep this code here in case
            #scale_x = abs(aff[0] + aff[3] + aff[6])
            #scale_y = abs(aff[1] + aff[4] + aff[7])
            #scale_z = abs(aff[2] + aff[5] + aff[8])
            #avg_scale = (scale_x + scale_y + scale_z) / 3.0;
            #
            # use determinant for scaling for now as it's most simple
            scale = det_scale
        for seg in self.compartment_list:
            x = seg[NODE_X] * aff[0] + seg[NODE_Y] * \
                aff[1] + seg[NODE_Z] * aff[2] + aff[9]
            y = seg[NODE_X] * aff[3] + seg[NODE_Y] * \
                aff[4] + seg[NODE_Z] * aff[5] + aff[10]
            z = seg[NODE_X] * aff[6] + seg[NODE_Y] * \
                aff[7] + seg[NODE_Z] * aff[8] + aff[11]
            seg[NODE_X] = x
            seg[NODE_Y] = y
            seg[NODE_Z] = z
            seg[NODE_R] *= scale

    def _separate_trees(self):
        """
        construct list of independent trees (each tree has a root of -1)
        """
        trees = []
        # reset each node's tree ID to indicate that it's not assigned
        for seg in self.compartment_list:
            seg[NODE_TREE_ID] = -1
        # construct trees for each node
        # if a node is adjacent an existing tree, merge to it
        # if a node is adjacent multiple trees, merge all
        for seg in self.compartment_list:
            # see what trees this node is adjacent to
            local_trees = []
            if seg[NODE_PN] >= 0 and self.compartment_list[seg[NODE_PN]][NODE_TREE_ID] >= 0:
                local_trees.append(self.compartment_list[
                                   seg[NODE_PN]][NODE_TREE_ID])
            for child_id in seg[NODE_CHILDREN]:
                child = self.compartment_list[child_id]
                if child[NODE_TREE_ID] >= 0:
                    local_trees.append(child[NODE_TREE_ID])
            # figure out which tree to put node into
            # if there are muliple possibilities, merge all of them
            if len(local_trees) == 0:
                tree_num = len(trees)  # create new tree
            elif len(local_trees) == 1:
                tree_num = local_trees[0]   # use existing tree
            elif len(local_trees) > 1:
                # this node is an intersection of multiple trees
                # merge all trees into the first one found
                tree_num = local_trees[0]
                for j in range(1, len(local_trees)):
                    dead_tree = local_trees[j]
                    trees[dead_tree] = []
                    for node in self.compartment_list:
                        if node[NODE_TREE_ID] == dead_tree:
                            node[NODE_TREE_ID] = tree_num
            # merge node into tree
            # ensure there's space
            while len(trees) <= tree_num:
                trees.append([])
            trees[tree_num].append(seg)
            seg[NODE_TREE_ID] = tree_num
        # consolidate tree lists into class's tree list object
        self._tree_list = []
        for tree in trees:
            if len(tree) > 0:
                self._tree_list.append(tree)
        # make soma's tree be the first tree, if soma present
        # this should be the case if the file is properly ordered, but
        #   don't assume that
        soma_tree = -1
        for seg in self.compartment_list:
            if seg[NODE_TYPE] == 1:
                soma_tree = seg[NODE_TREE_ID]
                break
        if soma_tree > 0:
            # swap soma tree for first tree in list
            tmp = self._tree_list[soma_tree]
            self._tree_list[soma_tree] = self._tree_list[0]
            self._tree_list[0] = tmp
        # reset node tree_id to correct tree number
        self._reset_tree_ids()

    def _reset_tree_ids(self):
        """
        reset each node's tree_id value to the correct tree number
        """
        for i in range(len(self._tree_list)):
            for j in range(len(self._tree_list[i])):
                self._tree_list[i][j][NODE_TREE_ID] = i

    def _check_consistency(self):
        """
        internal function -- don't publish in the docs
        TODO? print warning if unrecognized types are present
        Return value: number of errors detected in file
        """
        errs = 0
        # Make sure that the parents are of proper ID range
        n = self.num_nodes
        for seg in self.compartment_list:
            if seg[NODE_PN] >= 0:
                if seg[NODE_PN] >= n:
                    print("Parent for node %d is invalid (%d)" %
                          (seg[NODE_ID], seg[NODE_PN]))
                    errs += 1
        # make sure that each tree has exactly one root
        for i in range(self.num_trees):
            tree = self.tree(i)
            root = -1
            for j in range(len(tree)):
                if tree[j][NODE_PN] == -1:
                    if root >= 0:
                        print("Too many roots in tree %d" % i)
                        errs += 1
                    root = j
            if root == -1:
                print("No root present in tree %d" % i)
                errs += 1
        # make sure each axon has at most one root
        # find type boundaries. at each axon boundary, walk back up
        #   tree to root and make sure another axon segment not
        #   encountered
        adoptees = self._find_type_boundary()
        for child in adoptees:
            if child[NODE_TYPE] == Morphology.AXON:
                par_id = child[NODE_PN]
                while par_id >= 0:
                    par = self.compartment_list[par_id]
                    if par[NODE_TYPE] == Morphology.AXON:
                        print("Branch has multiple axon roots")
                        print(child)
                        print(par)
                        errs += 1
                        break
                    par_id = par[NODE_PN]
        if errs > 0:
            print("Failed consistency check: %d errors encountered" % errs)
        return errs

    def _find_type_boundary(self):
        """
        return a list of segments who have parents that are a different type
        """
        adoptees = []
        for node in self.compartment_list:
            par = self.parent_of(node)
            if par is None:
                continue
            if node[NODE_TYPE] != par[NODE_TYPE]:
                adoptees.append(node)
        return adoptees

    # remove tree from swc's "forest"
    def delete_tree(self, n):
        """ Delete tree, and all of its compartments, from the morphology.

        Parameters
        ----------
        n: Integer
            The tree number to delete
        """
        if n < 0:
            return
        if n >= self.num_trees:
            print("Error -- attempted to delete non-existing tree (%d)" % n)
            raise ValueError
        tree = self.tree(n)
        for i in range(len(tree)):
            self.compartment_list[tree[i][NODE_ID]] = None
        del self._tree_list[n]
        self._reconstruct()
        # reset node tree_id to correct tree number
        self._reset_tree_ids()

    def _print_all_nodes(self):
        """
        debugging function. prints all nodes
        """
        for node in self.compartment_list:
            print(node)

########################################################################
class Marker(dict):
    """ Simple dictionary class for handling reconstruction marker objects. """

    SPACING = [.1144, .1144, .28]

    CUT_DENDRITE = 10
    NO_RECONSTRUCTION = 20

    def __init__(self, *args, **kwargs):
        super(Marker, self).__init__(*args, **kwargs)

        # marker file x,y,z coordinates are offset by a single image-space
        # pixel
        self['x'] -= self.SPACING[0]
        self['y'] -= self.SPACING[1]
        self['z'] -= self.SPACING[2]


def read_marker_file(file_name):
    """ read in a marker file and return a list of dictionaries """

    with open(file_name, 'r') as f:
        rows = csv.DictReader((r for r in f if not r.startswith('#')),
                              fieldnames=['x', 'y', 'z', 'radius', 'shape', 'name', 'comment',
                                          'color_r', 'color_g', 'color_b'])

        return [Marker({'x': float(r['x']),
                        'y': float(r['y']),
                        'z': float(r['z']),
                        'name': int(r['name'])}) for r in rows]



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import allensdk.core.cell_types_cache as CTC
from allensdk.core.cell_types_cache import ReporterStatus as RS
import pytest
from pandas.core.frame import DataFrame
from allensdk.config import enable_console_log
from mock import MagicMock, patch, call, mock_open
from six.moves import builtins
import itertools as it
import allensdk.core.json_utilities as ju
import pandas.io.json as pj
import os


try:
    reload
except NameError:
    try:
        from importlib import reload
    except ImportError:
        from imp import reload


@pytest.fixture(scope='module', autouse=True)
@patch('allensdk.core.nwb_data_set.NwbDataSet')
def mock_nwb(nwb):
    reload(CTC)
    
    return nwb

_TEMPDIR = 'cell_types'
_MOCK_PATH = '/path/to/xyz.txt'

@pytest.fixture(scope="session", autouse=True)
def console_log():
    enable_console_log()

@pytest.fixture
def cell_id():
    cell_id = 480114344

    return cell_id


@pytest.fixture
def mock_csv():
    DataFrame.to_csv = MagicMock(name='to_csv')
    DataFrame.from_csv = MagicMock(name='from_csv')

    return None


@pytest.fixture
def mock_csv_stuff():
    stuff_json = [{ 'stuff': 'whatever'},
                  { 'stuff': 'nonsense'}]
    
    DataFrame.to_csv = MagicMock(name='to_csv')
    DataFrame.from_csv = MagicMock(name='from_csv',
                                   return_value=DataFrame(stuff_json))

    return None


@pytest.fixture
def mock_json():
    _msg = [{'stuff': 'nonsense'}]
    _pd_msg = DataFrame(_msg)

    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': _msg})
    ju.write = \
        MagicMock(name='write')

    ju.read = \
        MagicMock(name='read',
                  return_value=_msg)

    pj.read_json = \
        MagicMock(name='read_json',
                  return_value=_pd_msg)

    return None


@pytest.fixture
def cache_fixture():
    # Instantiate the CellTypesCache instance.  The manifest_file argument
    # tells it where to store the manifest, which is a JSON file that tracks
    # file paths.  If you supply a relative path (like this), it will go
    # into your current working directory
    ctc = CTC.CellTypesCache(
        manifest_file=os.path.join(_TEMPDIR,
        'cell_types_manifest.json'))

    ctc.get_cache_path = MagicMock(name='get_cache_path',
                                   return_value=_MOCK_PATH)
    return ctc


@pytest.fixture
def cache_fixture_cache_false():
    # Instantiate the CellTypesCache instance.  The manifest_file argument
    # tells it where to store the manifest, which is a JSON file that tracks
    # file paths.  If you supply a relative path (like this), it will go
    # into your current working directory
    ctc = CTC.CellTypesCache(cache=False,
        manifest_file=os.path.join(_TEMPDIR,
        'cell_types_manifest.json'))

    ctc.get_cache_path = MagicMock(name='get_cache_path',
                                   return_value=_MOCK_PATH)
    return ctc


@pytest.mark.parametrize('path_exists',
                         (False, True))
def test_sweep_data_with_api(cache_fixture,
                             mock_nwb,
                             path_exists):
    ctc = cache_fixture

    specimen_id = 464212183

    ephys_result = [{'ephys_result':
                     {'well_known_files': [
                      {'download_link': '/path/to/data.nwb' }]}}]

    # this saves the NWB file to 'cell_types/specimen_464212183/ephys.nwb'
    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.retrieve_file_over_http') as mock_http:
        with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
               MagicMock(name='model query',
                         return_value=ephys_result)) as query_mock:
            with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
                with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs') as mkd:
                    mock_nwb.reset_mock()
                    _ = ctc.get_ephys_data(specimen_id, _MOCK_PATH)

    assert ope.called
    if path_exists:
        mock_nwb.assert_called_once_with(_MOCK_PATH)
        assert not query_mock.called
        assert not mkd.called
    else:
        # both levels of cacheable methods check if the directory exists.
        assert mkd.call_args_list == [call(_MOCK_PATH)]
        assert query_mock.called
        mock_http.assert_called_once_with('http://api.brain-map.org/path/to/data.nwb',
                                          _MOCK_PATH)


def test_sweep_data_exception(cache_fixture):
    ctc = cache_fixture

    specimen_id = 464212183

    ephys_result = [{'ephys_result':
                     {'well_known_files': [] }}]

    with pytest.raises(Exception) as exc:
        with patch('allensdk.api.queries.cell_types_api.CellTypesApi.retrieve_file_over_http') as mock_http:
            with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
                   MagicMock(name='model query',
                             return_value=ephys_result)) as query_mock:
                with patch('os.path.exists', MagicMock(return_value=False)) as ope:
                    with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                        with patch('allensdk.core.cell_types_cache.NwbDataSet') as nwb:
                            _ = ctc.get_ephys_data(specimen_id)
    
    assert 'has no ephys data' in str(exc.value)


@pytest.mark.parametrize('path_exists,morph_flag,recon_flag,statuses',
                         it.product((False, True),
                                    (False, True),
                                    (False, True),
                                    (RS.POSITIVE, ['list', 'of', 'statuses'])))
def test_get_cells(cache_fixture,
                   path_exists,
                   morph_flag,
                   recon_flag,
                   statuses):
    ctc = cache_fixture
    # this downloads metadata for all cells with morphology images
    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.list_cells',
               MagicMock(return_value=['mock_cells_from_server'])) as list_cells_mock:
        with patch('allensdk.api.queries.cell_types_api.CellTypesApi.filter_cells',
                   MagicMock(return_value=['mock_cells'])) as filter_cells_mock:
            with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
                with patch('allensdk.core.json_utilities.read',
                           return_value=['mock_data']) as ju_read:
                    with patch('allensdk.core.json_utilities.write') as ju_write:
                        cells = ctc.get_cells(require_morphology=morph_flag,
                                              require_reconstruction=recon_flag,
                                              reporter_status=statuses)

    assert len(cells) > 0

    if path_exists:
        ju_read.assert_called_once_with(_MOCK_PATH)
        expected_cells = ['mock_data']
    else:
        list_cells_mock.assert_called_once_with(False, False)
        expected_cells = ['mock_cells_from_server']

    if (statuses == RS.POSITIVE):
        expected_status = [statuses]
    else:
        expected_status = statuses

    filter_cells_mock.assert_called_once_with(expected_cells,
                                              morph_flag,
                                              recon_flag,
                                              expected_status)


@pytest.mark.parametrize('path_exists,morph_flag,recon_flag,statuses',
                         it.product((False, True),
                                    (False, True),
                                    (False, True),
                                    (RS.POSITIVE, ['list', 'of', 'statuses'])))
def test_get_cells_with_api(cache_fixture,
                            path_exists,
                            morph_flag,
                            recon_flag,
                            statuses):
    ctc = cache_fixture

    # note, this is only a mock for coverage,
    # and has not a lot of relation to the actual data form
    sweeps = [1, 2, 3]
    return_dicts = [{'sweep_number': x,
                     'tags': ['what - ever'],
                     'neuron_reconstructions' : [],
                     'data_sets': [],
                     'reporter_status': 'whatever',
                     'has_morphology': False,
                     'has_reconstruction': False,
                     'donor': { 'transgenic_lines': [{'transgenic_line_type_name': 'driver',
                                                      'name': 'harold'}]},
                     'cell_reporter': {'name': 'tired'},
                     'specimen_tags': [{'name': 'a - b',
                                        'value': 123}]} for x \
                     in sweeps]

    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
               MagicMock(name='model query',
                         return_value=return_dicts)) as query_mock:
        with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
            with patch('allensdk.core.json_utilities.read',
                       return_value=return_dicts) as ju_read:
                with patch('allensdk.core.json_utilities.write') as ju_write:
                    with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                        cells = ctc.get_cells(require_morphology=morph_flag,
                                              require_reconstruction=recon_flag,
                                              reporter_status=statuses)

    if path_exists:
        ju_read.assert_called_once_with(_MOCK_PATH)
    else:
        assert ju_write.called

@pytest.mark.parametrize('path_exists',
                         (False, True))
def test_get_reconstruction(cache_fixture,
                               cell_id,
                               path_exists):
    ctc = cache_fixture

    save_recon = \
        'allensdk.api.queries.cell_types_api.CellTypesApi.save_reconstruction'
    with patch(save_recon) as save_recon_mock:
        with patch('allensdk.core.swc.read_swc') as read_swc_mock:
            # download and open an SWC file
            _ = ctc.get_reconstruction(cell_id)

    if path_exists is False:
        save_recon_mock.assert_called_once_with(cell_id,
                                                _MOCK_PATH)

    read_swc_mock.assert_called_once_with(_MOCK_PATH)


@pytest.mark.parametrize('path_exists',
                         (False, True))
def test_get_reconstruction_with_api(cache_fixture,
                                     mock_csv,
                                     cell_id,
                                     path_exists):
    ctc = cache_fixture

    reconstruction_data = [{'neuron_reconstructions': [
                            {'well_known_files': [
                             {'download_link': 'http://example.org'}]}]}]

    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.retrieve_file_over_http') as mock_http:
        with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
                   MagicMock(name='model query',
                             return_value=reconstruction_data)) as query_mock:
            with patch('allensdk.core.swc.read_swc') as read_swc_mock:
                with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
                    with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                        _ = ctc.get_reconstruction(cell_id)

    if path_exists:
        read_swc_mock.assert_called_once_with(_MOCK_PATH)
    else:
        assert query_mock.called


def test_get_reconstruction_exception(cache_fixture,
                                      mock_csv,
                                      cell_id):
    ctc = cache_fixture

    reconstruction_data = [{'neuron_reconstructions': [
                            {'well_known_files': None}]}]

    with pytest.raises(Exception) as exc:
        with patch('allensdk.api.queries.cell_types_api.CellTypesApi.retrieve_file_over_http') as mock_http:
            with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
                       MagicMock(name='model query',
                                 return_value=reconstruction_data)) as query_mock:
                with patch('allensdk.core.swc.read_swc') as read_swc_mock:
                    with patch('os.path.exists', MagicMock(return_value=False)) as ope:
                        with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                            _ = ctc.get_reconstruction(cell_id)

    assert 'has no reconstruction' in str(exc.value)


@pytest.mark.parametrize('path_exists,lookup_error',
                         it.product((False, True),
                                    (False, True)))
def test_get_reconstruction_markers(cache_fixture,
                                    cell_id,
                                    path_exists,
                                    lookup_error):
    ctc = cache_fixture

    if lookup_error:
        def lookup(i, n):
            raise(LookupError('mock lookup error'))
    else:
        def lookup(i, n):
            return

    save_recon_marker = \
        'allensdk.api.queries.cell_types_api.CellTypesApi.save_reconstruction_markers'

    # download and open a marker file
    with patch(save_recon_marker,
               MagicMock(side_effect=lookup)) as save_recon_markers_mock:
        with patch('allensdk.core.swc.read_marker_file') as read_marker_mock:
            _ = ctc.get_reconstruction_markers(cell_id)

    if path_exists is False:
        save_recon_markers_mock.assert_called_once_with(cell_id,
                                                       _MOCK_PATH)

    if lookup_error:
        assert not read_marker_mock.called
    else:
        read_marker_mock.assert_called_once_with(_MOCK_PATH)


@pytest.mark.parametrize('path_exists,lookup_error',
                         it.product((False, True),
                                    (False, True)))
def test_get_reconstruction_markers_with_api(cache_fixture,
                                             cell_id,
                                             path_exists,
                                             lookup_error):
    ctc = cache_fixture

    reconstruction_data = [{'neuron_reconstructions': [
                            {'well_known_files': [
                             {'download_link': '/mock/path_to_file'}]}]}]

    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.retrieve_file_over_http') as mock_http:
        with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
                   MagicMock(name='model query',
                             return_value=reconstruction_data)) as query_mock:
            with patch('allensdk.core.swc.read_marker_file') as marker_mock:
                with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
                    with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                        _ = ctc.get_reconstruction_markers(cell_id)

    if path_exists:
        assert marker_mock.called
    else:
        mock_http.assert_called_once_with('http://api.brain-map.org/mock/path_to_file',
                                          _MOCK_PATH)


def test_get_reconstruction_markers_exception(cache_fixture,
                                             cell_id):
    ctc = cache_fixture

    reconstruction_data = [{'neuron_reconstructions': [
                            {'well_known_files': []}]}]

    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.retrieve_file_over_http') as mock_http:
        with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
                   MagicMock(name='model query',
                             return_value=reconstruction_data)) as query_mock:
            with patch('allensdk.core.swc.read_marker_file') as marker_mock:
                with patch('os.path.exists', MagicMock(return_value=False)) as ope:
                    with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                        markers = ctc.get_reconstruction_markers(cell_id)

                        assert len(markers) == 0


@pytest.mark.parametrize('dataframe',
                         (False, True))
def test_get_ephys_features(cache_fixture,
                            dataframe):
    ctc = cache_fixture

    api_get_ephys_features = \
        'allensdk.api.queries.cell_types_api.CellTypesApi.get_ephys_features'

    with patch(api_get_ephys_features) as api_get_ephys_features_mock:
        # download all electrophysiology features for all cells
        _ = ctc.get_ephys_features(dataframe=dataframe)

    assert api_get_ephys_features_mock.called


@pytest.mark.parametrize('df,path_exists',
                         it.product((False,True),
                                    (False,True)))
def test_get_ephys_features_with_api(cache_fixture,
                                     mock_csv,
                                     df,
                                     path_exists):
    ctc = cache_fixture

    mock_data = [{'lorem': 1,
                  'ipsum': 2 },
                 {'lorem': 3,
                  'ipsum': 4 }]

    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
               MagicMock(name='model query',
                         return_value=mock_data)) as query_mock:
        with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
            with patch(builtins.__name__ + '.open',
                       mock_open(),
                       create=True) as open_mock:
                with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs') as mkd:
                    _ = ctc.get_ephys_features(dataframe=df)

    if path_exists:
        DataFrame.from_csv.assert_called_once_with(_MOCK_PATH)
    else:
        mkd.assert_called_once_with(_MOCK_PATH)
        assert query_mock.called


@pytest.mark.parametrize('path_exists,df',
                         it.product((False, True),
                                    (False, True)))
def test_get_morphology_features(cache_fixture,
                                 mock_csv_stuff,
                                 path_exists,
                                 df):
    ctc = cache_fixture

    json_data = [{ 'stuff': 'whatever'},
                 { 'stuff': 'nonsense'}]
    
    with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
        with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs') as mkd:
            with patch(builtins.__name__ + '.open',
                       mock_open(),
                       create=True) as open_mock:
                with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
                           MagicMock(name='model query',
                                     return_value=json_data)) as query_mock:
                    data = ctc.get_morphology_features(df,
                                                       _MOCK_PATH)

    if df:
        assert ('stuff' in data) == True
    else:
        assert all(['stuff' in f for f in data])

    
    if path_exists:
        if df:
            DataFrame.from_csv.assert_called_once_with(_MOCK_PATH)
        else:
            assert True
        assert not mkd.called
    else:
        assert query_mock.called
        assert mkd.called


@pytest.mark.parametrize('path_exists',
                         (False, True))
def test_get_ephys_sweeps(cache_fixture,
                          path_exists):
    ctc = cache_fixture

    cell_id = 464212183

    get_ephys_sweeps = \
        'allensdk.api.queries.cell_types_api.CellTypesApi.get_ephys_sweeps'
    with patch(get_ephys_sweeps) as get_ephys_sweeps_mock:
        with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
            with patch('allensdk.core.json_utilities.read',
                       return_value=['mock_data']) as ju_read:
                with patch('allensdk.core.json_utilities.write') as ju_write:
                    _ = ctc.get_ephys_sweeps(cell_id)

    if path_exists:
        assert ju_read.called_once_with(_MOCK_PATH)
    else:
        assert get_ephys_sweeps_mock.called_once_with(cell_id)


@pytest.mark.parametrize('path_exists',
                         (False, True))
def test_get_ephys_sweeps_with_api(cache_fixture,
                                   path_exists):
    ctc = cache_fixture

    cell_id = 464212183
    sweeps = [1, 2, 3]
    return_dicts = [{'sweep_number': x} for x in sweeps]

    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
               MagicMock(name='model query',
                         return_value=return_dicts)) as query_mock:
        with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
            with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                with patch('allensdk.core.json_utilities.read',
                           return_value=['mock_data']) as ju_read:
                    with patch('allensdk.core.json_utilities.write') as ju_write:
                        _ = ctc.get_ephys_sweeps(cell_id)

    # read will be called regardless
    assert ju_read.called_once_with(_MOCK_PATH)

    if path_exists:
        assert not query_mock.called
    else:
        assert query_mock.called


@pytest.mark.parametrize('path_exists,require_reconstruction',
                         it.product((False, True),
                                    (False, True)))
@patch('pandas.DataFrame.merge')
def test_get_all_features(mock_merge,
                           cache_fixture,
                           mock_csv_stuff,
                           path_exists,
                           require_reconstruction):
    ctc = cache_fixture

    sweeps = [1, 2, 3]
    return_dicts = [{'sweep_number': x,
                     'tags': 'whatever'} for x in sweeps]

    with patch('allensdk.api.queries.cell_types_api.CellTypesApi.model_query',
               MagicMock(name='model query',
                         return_value=return_dicts)) as query_mock:
            with patch('os.path.exists', MagicMock(return_value=path_exists)) as ope:
                with patch('allensdk.config.manifest.Manifest.safe_make_parent_dirs'):
                    with patch('allensdk.core.json_utilities.read',
                               return_value=return_dicts) as ju_read:
                        with patch(builtins.__name__ + '.open',
                                   mock_open(),
                                   create=True) as open_mock:
                            with patch('allensdk.core.json_utilities.write') as ju_write:
                                _ = ctc.get_all_features(
                                    require_reconstruction=require_reconstruction)

    if path_exists:
        assert DataFrame.from_csv.called
    else:
        assert query_mock.called
    
    assert mock_merge.called


def test_build_manifest(cache_fixture):
    ctc = cache_fixture
    
    mb_mock = MagicMock(name='manifest builder')

    with patch('allensdk.core.cell_types_cache.ManifestBuilder',
               return_value=mb_mock):
        ctc.build_manifest('test_manifest.json')

    assert mb_mock.add_path.call_count == 8
    mb_mock.write_json_file.assert_called_once_with('test_manifest.json')



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import logging
import six
from . import ephys_extractor as efex
from . import ephys_features as ft

HERO_MIN_AMP_OFFSET = 39.0
HERO_MAX_AMP_OFFSET = 61.0

SHORT_SQUARE_TYPES = ["Short Square",
                      "Short Square - Triple",
                      "Short Square - Hold -60mv",
                      "Short Square - Hold -70mv",
                      "Short Square - Hold -80mv"]

SHORT_SQUARE_THRESH_FRAC_FLOOR = 0.1

MEAN_FEATURES = [ "upstroke_downstroke_ratio", "peak_v", "peak_t", "trough_v", "trough_t",
                  "fast_trough_v", "fast_trough_t", "slow_trough_v", "slow_trough_t",
                  "threshold_v", "threshold_i", "threshold_t", "peak_v", "peak_t" ]


def extract_sweep_features(data_set, sweeps_by_type):
    # extract sweep-level features
    sweep_features = {}

    for stimulus_type, sweep_numbers in six.iteritems(sweeps_by_type):
        logging.debug("%s:%s" % (stimulus_type, ','.join(map(str, sweep_numbers))))

        if stimulus_type == "Short Square - Triple":
            tmp_ext = efex.extractor_for_nwb_sweeps(data_set, sweep_numbers)
            t_set = [s.t for s in tmp_ext.sweeps()]
            v_set = [s.v for s in tmp_ext.sweeps()]

            # IT-14530
            # triple-sweeps to use different window
            win_start = efex.SHORT_SQUARE_TRIPLE_WINDOW_START
            win_end = efex.SHORT_SQUARE_TRIPLE_WINDOW_END
            cutoff, thresh_frac = ft.estimate_adjusted_detection_parameters(
                                    v_set, t_set, win_start, win_end)
            thresh_frac = max(SHORT_SQUARE_THRESH_FRAC_FLOOR, thresh_frac)

            fex = efex.extractor_for_nwb_sweeps(data_set, sweep_numbers,
                                    dv_cutoff=cutoff, thresh_frac=thresh_frac)
        elif stimulus_type in SHORT_SQUARE_TYPES:
            tmp_ext = efex.extractor_for_nwb_sweeps(data_set, sweep_numbers)
            t_set = [s.t for s in tmp_ext.sweeps()]
            v_set = [s.v for s in tmp_ext.sweeps()]

            win_start = efex.SHORT_SQUARES_WINDOW_START
            win_end = efex.SHORT_SQUARES_WINDOW_END
            cutoff, thresh_frac = ft.estimate_adjusted_detection_parameters(
                                     v_set, t_set, win_start, win_end)
            thresh_frac = max(SHORT_SQUARE_THRESH_FRAC_FLOOR, thresh_frac)

            fex = efex.extractor_for_nwb_sweeps(data_set, sweep_numbers,
                                                dv_cutoff=cutoff, thresh_frac=thresh_frac)
        else:
            fex = efex.extractor_for_nwb_sweeps(data_set, sweep_numbers)

        fex.process_spikes()

        sweep_features.update({ f.id:f.as_dict() for f in fex.sweeps() })

    return sweep_features

# if subthreshold minimum amplitude is known (e.g., for human cells) then
#   specify it. otherwise the default value will be used
def extract_cell_features(data_set,
                          ramp_sweep_numbers,
                          short_square_sweep_numbers,
                          long_square_sweep_numbers,
                          subthresh_min_amp = None):

    if subthresh_min_amp is None:
        fex = efex.cell_extractor_for_nwb(data_set,
                                          ramp_sweep_numbers,
                                          short_square_sweep_numbers,
                                          long_square_sweep_numbers)
    else:
        fex = efex.cell_extractor_for_nwb(data_set,
                                          ramp_sweep_numbers,
                                          short_square_sweep_numbers,
                                          long_square_sweep_numbers,
                                          subthresh_min_amp)

    fex.process()

    cell_features = fex.as_dict()

    # find hero sweep
    rheo_amp = cell_features['long_squares']['rheobase_i']
    hero_min, hero_max = rheo_amp + HERO_MIN_AMP_OFFSET, rheo_amp + HERO_MAX_AMP_OFFSET
    hero_amp = float("inf")
    hero_sweep = None
    for sweep in fex.long_squares_features("spiking").sweeps():
        nspikes = len(sweep.spikes())
        amp = sweep.sweep_feature("stim_amp")

        if nspikes > 0 and amp > hero_min and amp < hero_max and amp < hero_amp:
            hero_amp = amp
            hero_sweep = sweep

    if hero_sweep:
        adapt = hero_sweep.sweep_feature("adapt")
        latency = hero_sweep.sweep_feature("latency")
        mean_isi = hero_sweep.sweep_feature("mean_isi")
    else:
        raise ft.FeatureError("Could not find hero sweep.")

    # find the mean features of the first spike for the ramps and short squares
    ramps_ms0 = mean_features_spike_zero(fex.ramps_features().sweeps())
    ss_ms0 = mean_features_spike_zero(fex.short_squares_features().sweeps())

    # compute baseline from all long square sweeps
    v_baseline = np.mean(fex.long_squares_features().sweep_features('v_baseline'))

    cell_features['long_squares']['v_baseline'] = v_baseline
    cell_features['long_squares']['hero_sweep'] = hero_sweep.as_dict() if hero_sweep else None
    cell_features["ramps"]["mean_spike_0"] = ramps_ms0
    cell_features["short_squares"]["mean_spike_0"] = ss_ms0

    return cell_features

def mean_features_spike_zero(sweeps):
    """ Compute mean feature values for the first spike in list of extractors """

    output = {}
    for mf in MEAN_FEATURES:
        mfd = [ sweep.spikes()[0][mf] for sweep in sweeps if sweep.sweep_feature("avg_rate") > 0 ]
        output[mf] = np.mean(mfd)
    return output

def get_stim_characteristics(i, t, no_test_pulse=False):
    '''
    Identify the start time, duration, amplitude, start index, and
    end index of a general stimulus.
    This assumes that there is a test pulse followed by the stimulus square.
    '''

    di = np.diff(i)
    diff_idx = np.flatnonzero(di != 0)

    if len(diff_idx) == 0:
        return (None, None, 0.0, None, None)

    # skip the first up/down
    idx = 0 if no_test_pulse else 1

    # shift by one to compensate for diff()
    start_idx = diff_idx[idx] + 1
    end_idx = diff_idx[-1] + 1

    stim_start = float(t[start_idx])
    stim_dur = float(t[end_idx] - t[start_idx])
    stim_amp = float(i[start_idx])

    return (stim_start, stim_dur, stim_amp, start_idx, end_idx)

def get_ramp_stim_characteristics(i, t):
    ''' Identify the start time and start index of a ramp sweep. '''

    # Assumes that there is a test pulse followed by the stimulus ramp
    di = np.diff(i)
    up_idx = np.flatnonzero(di > 0)

    start_idx = up_idx[1] + 1 # shift by one to compensate for diff()
    return (t[start_idx], start_idx)

def get_square_stim_characteristics(i, t, no_test_pulse=False):
    '''
    Identify the start time, duration, amplitude, start index, and
    end index of a square stimulus.
    This assumes that there is a test pulse followed by the stimulus square.
    '''

    di = np.diff(i)
    up_idx = np.flatnonzero(di > 0)
    down_idx = np.flatnonzero(di < 0)

    idx = 0 if no_test_pulse else 1

    # second square is the stimulus
    if up_idx[idx] < down_idx[idx]: # positive square
        start_idx = up_idx[idx] + 1 # shift by one to compensate for diff()
        end_idx = down_idx[idx] + 1
    else: # negative square
        start_idx = down_idx[idx] + 1
        end_idx = up_idx[idx] + 1

    stim_start = float(t[start_idx])
    stim_dur = float(t[end_idx] - t[start_idx])
    stim_amp = float(i[start_idx])

    return (stim_start, stim_dur, stim_amp, start_idx, end_idx)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from .rma_api import RmaApi
from .grid_data_api import GridDataApi
from ..cache import cacheable, Cache
import numpy as np
import nrrd
import six


class MouseConnectivityApi(RmaApi):
    '''
    HTTP Client for the Allen Mouse Brain Connectivity Atlas.

    See: `Mouse Connectivity API <http://help.brain-map.org/display/mouseconnectivity/API>`_
    '''
    PRODUCT_IDS = [5, 31]

    def __init__(self, base_uri=None):
        super(MouseConnectivityApi, self).__init__(base_uri)

    AVERAGE_TEMPLATE = 'average_template'
    ARA_NISSL = 'ara_nissl'
    MOUSE_2011 = 'annotation/mouse_2011'
    DEVMOUSE_2012 = 'annotation/devmouse_2012'
    CCF_2015 = 'annotation/ccf_2015'
    CCF_2016 = 'annotation/ccf_2016'
    CCF_VERSION_DEFAULT = CCF_2016

    VOXEL_RESOLUTION_10_MICRONS = 10
    VOXEL_RESOLUTION_25_MICRONS = 25
    VOXEL_RESOLUTION_50_MICRONS = 50
    VOXEL_RESOLUTION_100_MICRONS = 100

    @cacheable(strategy='create',
               reader=nrrd.read,
               pathfinder=Cache.pathfinder(file_name_position=3,
                                           path_keyword='file_name'))
    def download_annotation_volume(self,
                                   ccf_version,
                                   resolution,
                                   file_name):
        '''
        Download the annotation volume at a particular resolution.

        Parameters
        ----------
        ccf_version: string
            MouseConnectivityApi.CCF_2016 (default) or MouseConnectivityApi.CCF_2015
        resolution: int
            Desired resolution to download in microns.
            Must be 10, 25, 50, or 100.
        file_name: string
            Where to save the annotation volume.
        
        Note: the parameters must be used as positional parameters, not keywords
        '''

        if ccf_version is None:
            ccf_version = MouseConnectivityApi.CCF_VERSION_DEFAULT

        self.download_volumetric_data(ccf_version,
                                      'annotation_%d.nrrd' % resolution, 
                                      save_file_path=file_name)

    @cacheable(strategy='create',
               reader=nrrd.read,
               pathfinder=Cache.pathfinder(file_name_position=2,
                                           path_keyword='file_name'))
    def download_template_volume(self, resolution, file_name):
        '''
        Download the registration template volume at a particular resolution.

        Parameters
        ----------

        resolution: int
            Desired resolution to download in microns.  Must be 10, 25, 50, or 100.

        file_name: string
            Where to save the registration template volume.
        '''
        self.download_volumetric_data(MouseConnectivityApi.AVERAGE_TEMPLATE,
                                      'average_template_%d.nrrd' % resolution, 
                                      save_file_path=file_name)


    @cacheable()
    def get_experiments(self,
                        structure_ids,
                        **kwargs):
        '''
        Fetch experiment metadata from the Mouse Brain Connectivity Atlas.

        Parameters
        ----------
        structure_ids : integer or list, optional
            injection structure

        Returns
        -------
        url : string
            The constructed URL
        '''
        criteria_list = ['[failed$eqfalse]',
                         'products[id$in%s]' % (','.join(str(i) for i in MouseConnectivityApi.PRODUCT_IDS))]

        if structure_ids is not None:
            if type(structure_ids) is not list:
                structure_ids = [structure_ids]
            criteria_list.append('[id$in%s]' % ','.join(str(i)
                                                        for i in structure_ids))

        criteria_string = ','.join(criteria_list)

        return self.model_query('SectionDataSet',
                                criteria=criteria_string,
                                **kwargs)

    @cacheable()
    def get_manual_injection_summary(self, experiment_id):
        ''' Retrieve manual injection summary. '''

        criteria = '[id$in%d]' % (experiment_id)

        include = ['specimen(donor(transgenic_mouse(transgenic_lines)),',
                   'injections(structure,age)),',
                   'equalization,products']

        only = ['id',
                'failed',
                'storage_directory',
                'red_lower',
                'red_upper',
                'green_lower',
                'green_upper',
                'blue_lower',
                'blue_upper',
                'products.id',
                'specimen_id',
                'structure_id',
                'reference_space_id',
                'primary_injection_structure_id',
                'registration_point',
                'coordinates_ap',
                'coordinates_dv',
                'coordinates_ml',
                'angle',
                'sex',
                'strain',
                'injection_materials',
                'acronym',
                'structures.name',
                'days',
                'transgenic_mice.name',
                'transgenic_lines.name',
                'transgenic_lines.description',
                'transgenic_lines.id',
                'donors.id']

        return self.model_query('SectionDataSet',
                                criteria=criteria,
                                include=include,
                                only=only)

    @cacheable()
    def get_experiment_detail(self, experiment_id):
        '''Retrieve the experiments data.'''

        criteria = '[id$eq%d]' % (experiment_id)
        include = ['specimen(stereotaxic_injections(primary_injection_structure,structures,stereotaxic_injection_coordinates)),',
                   'equalization,',
                   'sub_images']
        order = ["'sub_images.section_number$asc'"]

        return self.model_query('SectionDataSet',
                                criteria=criteria,
                                include=include,
                                order=order)

    @cacheable()
    def get_projection_image_info(self,
                                  experiment_id,
                                  section_number):
        '''Fetch meta-information of one projection image.

        Parameters
        ----------
        experiment_id : integer

        section_number : integer

        Notes
        -----
        See: image examples under
        `Experimental Overview and Metadata <http://help.brain-map.org/display/mouseconnectivity/API##API-ExperimentalOverviewandMetadata>`_
        for additional documentation.
        Download the image using :py:meth:`allensdk.api.queries.image_download_api.ImageDownloadApi.download_section_image`
        '''

        criteria = '[id$eq%d]' % (experiment_id)
        include = ['equalization,sub_images[section_number$eq%d]' %
                   (section_number)]

        return self.model_query('SectionDataSet',
                                criteria=criteria,
                                include=include)

    def build_volumetric_data_download_url(self,
                                           data_path,
                                           file_name,
                                           voxel_resolution=None,
                                           release=None,
                                           coordinate_framework=None):
        '''Construct url to download 3D reference model in NRRD format.

        Parameters
        ----------
        data_path : string
            'average_template', 'ara_nissl', 'annotation/ccf_2015', 'annotation/mouse_2011', or 'annotation/devmouse_2012'
        voxel_resolution : int
            10, 25, 50 or 100
        coordinate_framework : string
            'mouse_ccf' (default) or 'mouse_annotation'

        Notes
        -----
        See: `3-D Reference Models <http://help.brain-map.org/display/mouseconnectivity/API#API-3DReferenceModels>`_
        for additional documentation.
        '''

        if voxel_resolution is None:
            voxel_resolution = MouseConnectivityApi.VOXEL_RESOLUTION_10_MICRONS

        if release is None:
            release = 'current-release'

        if coordinate_framework is None:
            coordinate_framework = 'mouse_ccf'

        url = ''.join([self.informatics_archive_endpoint,
                       '/%s/%s/' % (release, coordinate_framework),
                       data_path,
                       '/',
                       file_name])

        return url


    def download_volumetric_data(self,
                                 data_path,
                                 file_name,
                                 voxel_resolution=None,
                                 save_file_path=None,
                                 release=None,
                                 coordinate_framework=None):
        '''Download 3D reference model in NRRD format.

        Parameters
        ----------
        data_path : string
            'average_template', 'ara_nissl', 'annotation/ccf_2015', 'annotation/mouse_2011', or 'annotation/devmouse_2012'
        file_name : string
            server-side file name. 'annotation_10.nrrd' for example.
        voxel_resolution : int
            10, 25, 50 or 100
        coordinate_framework : string
            'mouse_ccf' (default) or 'mouse_annotation'

        Notes
        -----
        See: `3-D Reference Models <http://help.brain-map.org/display/mouseconnectivity/API#API-3DReferenceModels>`_
        for additional documentation.
        '''
        url = self.build_volumetric_data_download_url(data_path,
                                                      file_name,
                                                      voxel_resolution,
                                                      release,
                                                      coordinate_framework)

        if save_file_path is None:
            save_file_path = file_name

        if save_file_path is None:
            save_file_path = 'volumetric_data.nrrd'

        self.retrieve_file_over_http(url, save_file_path)
        

    def download_reference_aligned_image_channel_volumes(self,
                                                         data_set_id,
                                                         save_file_path=None):
        '''
        Returns
        -------
            The well known file is downloaded
        '''
        well_known_file_url = self.get_reference_aligned_image_channel_volumes_url(
            data_set_id)

        if save_file_path is None:
            save_file_path = str(data_set_id) + '.zip'

        self.retrieve_file_over_http(well_known_file_url, save_file_path)

    def build_reference_aligned_image_channel_volumes_url(self,
                                                          data_set_id):
        '''Construct url to download the red, green, and blue channels
        aligned to the 25um adult mouse brain reference space volume.

        Parameters
        ----------
        data_set_id : integerallensdk.api.queries
            aka attachable_id

        Notes
        -----
        See: `Reference-aligned Image Channel Volumes <http://help.brain-map.org/display/mouseconnectivity/API#API-ReferencealignedImageChannelVolumes>`_
        for additional documentation.
        '''

        criteria = ['well_known_file_type',
                    "[name$eq'ImagesResampledTo25MicronARA']",
                    "[attachable_id$eq%d]" % (data_set_id)]

        model_stage = self.model_stage('WellKnownFile',
                                       criteria=criteria)

        url = self.build_query_url([model_stage])

        return url

    def get_reference_aligned_image_channel_volumes_url(self,
                                                        data_set_id):
        '''Retrieve the download link for a specific data set.\

        Notes
        -----
        See `Reference-aligned Image Channel Volumes <http://help.brain-map.org/display/mouseconnectivity/API#API-ReferencealignedImageChannelVolumes>`_
        for additional documentation.
        '''
        download_link = self.do_query(self.build_reference_aligned_image_channel_volumes_url,
                                      lambda parsed_json: str(
                                          parsed_json['msg'][0]['download_link']),
                                      data_set_id)

        url = self.api_url + download_link

        return url

    def experiment_source_search(self, **kwargs):
        '''Search over the whole projection signal statistics dataset
        to find experiments with specific projection profiles.

        Parameters
        ----------
        injection_structures : list of integers or strings
            Integer Structure.id or String Structure.acronym.
        target_domain : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        injection_hemisphere : string, optional
            'right' or 'left', Defaults to both hemispheres.
        target_hemisphere : string, optional
            'right' or 'left', Defaults to both hemispheres.
        transgenic_lines : list of integers or strings, optional
             Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        injection_domain : list of integers or strings, optional
             Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        product_ids : list of integers, optional
            Integer Product.id
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.

        Notes
        -----
        See `Source Search <http://help.brain-map.org/display/mouseconnectivity/API#API-SourceSearch>`_,
        `Target Search <http://help.brain-map.org/display/mouseconnectivity/API#API-TargetSearch>`_,
        and
        `service::mouse_connectivity_injection_structure <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivityinjectionstructure>`_.

        '''
        tuples = [(k, v) for k, v in six.iteritems(kwargs)]
        return self.service_query('mouse_connectivity_injection_structure', parameters=tuples)

    def experiment_spatial_search(self, **kwargs):
        '''Displays all SectionDataSets
        with projection signal density >= 0.1 at the seed point.
        This service also returns the path
        along the most dense pixels from the seed point
        to the center of each injection site..

        Parameters
        ----------
        seed_point : list of floats
            The coordinates of a point in 3-D SectionDataSet space.
        transgenic_lines : list of integers or strings, optional
            Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        section_data_sets : list of integers, optional
            Ids to filter the results.
        injection_structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        product_ids : list of integers, optional
            Integer Product.id
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.

        Notes
        -----
        See `Spatial Search <http://help.brain-map.org/display/mouseconnectivity/API#API-SpatialSearch>`_
        and
        `service::mouse_connectivity_target_spatial <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivitytargetspatial>`_.

        '''

        tuples = [(k, v) for k, v in six.iteritems(kwargs)]
        return self.service_query('mouse_connectivity_target_spatial', parameters=tuples)

    def experiment_injection_coordinate_search(self, **kwargs):
        '''User specifies a seed location within the 3D reference space.
        The service returns a rank list of experiments
        by distance of its injection site to the specified seed location.

        Parameters
        ----------
        seed_point : list of floats
            The coordinates of a point in 3-D SectionDataSet space.
        transgenic_lines : list of integers or strings, optional
            Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        injection_structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        product_ids : list of integers, optional
            Integer Product.id
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.

        Notes
        -----
        See `Injection Coordinate Search <http://help.brain-map.org/display/mouseconnectivity/API#API-InjectionCoordinateSearch>`_
        and
        `service::mouse_connectivity_injection_coordinate <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivityinjectioncoordinate>`_.

        '''
        tuples = [(k, v) for k, v in six.iteritems(kwargs)]
        return self.service_query('mouse_connectivity_injection_coordinate', parameters=tuples)

    def experiment_correlation_search(self, **kwargs):
        '''Select a seed experiment and a domain over
        which the similarity comparison is to be made.


        Parameters
        ----------
        row : integer
            SectionDataSet.id to correlate against.
        structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        hemisphere : string, optional
            Use 'right' or 'left'. Defaults to both hemispheres.
        transgenic_lines : list of integers or strings, optional
            Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
        injection_structures : list of integers or strings, optional
            Integer Structure.id or String Structure.acronym.
        primary_structure_only : boolean, optional
        product_ids : list of integers, optional
            Integer Product.id
        start_row : integer, optional
            For paging purposes. Defaults to 0.
        num_rows : integer, optional
            For paging purposes. Defaults to 2000.

        Notes
        -----
        See `Correlation Search <http://help.brain-map.org/display/mouseconnectivity/API#API-CorrelationSearch>`_
        and
        `service::mouse_connectivity_correlation <http://help.brain-map.org/display/api/Connected+Services+and+Pipes#ConnectedServicesandPipes-service%3A%3Amouseconnectivitycorrelation>`_.

        '''
        tuples = sorted(six.iteritems(kwargs))
        return self.service_query('mouse_connectivity_correlation',
                                  parameters=tuples)

    @cacheable()
    def get_structure_unionizes(self,
                                experiment_ids,
                                is_injection=None,
                                structure_name=None,
                                structure_ids=None,
                                hemisphere_ids=None,
                                normalized_projection_volume_limit=None,
                                include=None,
                                debug=None,
                                order=None):

        experiment_filter = '[section_data_set_id$in%s]' %\
                            ','.join(str(i) for i in experiment_ids)

        if is_injection is True:
            is_injection_filter = '[is_injection$eqtrue]'
        elif is_injection is False:
            is_injection_filter = '[is_injection$eqfalse]'
        else:
            is_injection_filter = ''

        if normalized_projection_volume_limit is not None:
            volume_filter = '[normalized_projection_volume$gt%f]' %\
                            (normalized_projection_volume_limit)
        else:
            volume_filter = ''

        if hemisphere_ids is not None:
            hemisphere_filter = '[hemisphere_id$in%s]' %\
                ','.join(str(h) for h in hemisphere_ids)
        else:
            hemisphere_filter = ''

        if structure_name is not None:
            structure_filter = ",structure[name$eq'%s']" % (structure_name)
        elif structure_ids is not None:
            structure_filter = '[structure_id$in%s]' %\
                               ','.join(str(i) for i in structure_ids)
        else:
            structure_filter = ''

        return self.model_query(
            'ProjectionStructureUnionize',
            criteria=''.join([experiment_filter,
                              is_injection_filter,
                              volume_filter,
                              hemisphere_filter,
                              structure_filter]),
            include=include,
            order=order,
            num_rows='all',
            debug=debug,
            count=False)

    @cacheable(strategy='create', 
               pathfinder=Cache.pathfinder(file_name_position=1,
                                           path_keyword='path'))
    def download_injection_density(self, path, experiment_id, resolution):
        GridDataApi(base_uri=self.api_url).download_projection_grid_data(
            experiment_id, [GridDataApi.INJECTION_DENSITY], resolution, path)

    @cacheable(strategy='create', 
               pathfinder=Cache.pathfinder(file_name_position=1,
                                           path_keyword='path'))
    def download_projection_density(self, path, experiment_id, resolution):
        GridDataApi(base_uri=self.api_url).download_projection_grid_data(
            experiment_id, [GridDataApi.PROJECTION_DENSITY], resolution, path)

    @cacheable(strategy='create', 
               pathfinder=Cache.pathfinder(file_name_position=1,
                                           path_keyword='path'))
    def download_injection_fraction(self, path, experiment_id, resolution):
        GridDataApi(base_uri=self.api_url).download_projection_grid_data(
            experiment_id, [GridDataApi.INJECTION_FRACTION], resolution, path)

    @cacheable(strategy='create', 
               pathfinder=Cache.pathfinder(file_name_position=1,
                                           path_keyword='path'))
    def download_data_mask(self, path, experiment_id, resolution):
        GridDataApi(base_uri=self.api_url).download_projection_grid_data(
            experiment_id, [GridDataApi.DATA_MASK], resolution, path)

    def calculate_injection_centroid(self,
                                     injection_density,
                                     injection_fraction,
                                     resolution=25):
        '''
        Compute the centroid of an injection site.

        Parameters
        ----------

        injection_density: np.ndarray
            The injection density volume of an experiment

        injection_fraction: np.ndarray
            The injection fraction volume of an experiment

        '''

        # find all voxels with injection_fraction > 0
        injection_voxels = np.nonzero(injection_fraction)
        injection_density_computed = np.multiply(injection_density[injection_voxels],
                                                 injection_fraction[injection_voxels])
        sum_density = np.sum(injection_density_computed)

        # compute centroid in CCF coordinates
        if sum_density > 0:
            centroid = np.dot(injection_density_computed,
                              list(zip(*injection_voxels))) / sum_density * resolution
        else:
            centroid = None

        return centroid



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import Mock, patch
from allensdk.api.cache import cacheable, Cache
from allensdk.config.manifest import Manifest
import allensdk.core.json_utilities as ju
import pandas.io.json as pj
import pandas as pd

try:
    import StringIO
except:
    import io as StringIO

try:
    reload
except NameError:
    try:
        from importlib import reload
    except ImportError:
        from imp import reload


@pytest.fixture(scope='module', autouse=True)
def mock_imports():
    with patch('nrrd.read',
               Mock(name='nrrd_read_file_mcm',
                    return_value=('mock_annotation_data',
                                  'mock_annotation_image'))) as nrrd_read:
        import allensdk.api.queries.mouse_connectivity_api
        reload(allensdk.api.queries.mouse_connectivity_api)
        from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi as MCA

    return nrrd_read, MCA


@pytest.fixture
def cache():
    return Cache()


_msg = [{'whatever': True}]
_pd_msg = pd.DataFrame(_msg)
_csv_msg = pd.read_csv(StringIO.StringIO(""",whatever
0,True
"""))

@pytest.fixture
def mca(mock_imports):
    _, MCA = mock_imports
    
    ju.read_url_get = \
        Mock(name='read_url_get',
             return_value={'msg': _msg})
    ju.write = \
        Mock(name='write')

    ju.read = \
        Mock(name='read',
             return_value=_pd_msg)

    pj.read_json = \
        Mock(name='read_json',
             return_value=_pd_msg)

    pd.DataFrame.to_csv = \
        Mock(name='to_csv')

    pd.DataFrame.read_csv = \
        Mock(name='read_csv',
             return_value=_csv_msg)

    Manifest.safe_mkdir = Mock(name='safe_mkdir')

    mca = MCA()
    mca.retrieve_file_over_http = Mock(name='retrieve_file_over_http')
    
    return mca


@pytest.mark.parametrize("file_exists",
                         (True, False))
def test_file_download_lazy(mock_imports,
                            mca, 
                            cache,
                            file_exists):
    nrrd_read, MCA = mock_imports

    @cacheable(strategy='lazy',
               reader=nrrd_read,
               pathfinder=Cache.pathfinder(file_name_position=3,
                                           secondary_file_name_position=1))
    def download_volumetric_data(data_path,
                                 file_name,
                                 voxel_resolution=None,
                                 save_file_path=None,
                                 release=None,
                                 coordinate_framework=None):
        url = mca.build_volumetric_data_download_url(data_path,
                                                     file_name,
                                                     voxel_resolution,
                                                     release,
                                                     coordinate_framework)

        mca.retrieve_file_over_http(url, save_file_path)

    with patch('os.path.exists',
               Mock(name="os.path.exists",
                    return_value=file_exists)) as mkdir:
        nrrd_read.reset_mock()
        download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                 'annotation_10.nrrd',
                                 MCA.VOXEL_RESOLUTION_10_MICRONS,
                                 'volumetric.nrrd',
                                 MCA.CCF_2016,
                                 strategy='lazy')

    if file_exists:
        assert not mca.retrieve_file_over_http.called, 'server call not needed when file exists'
    else:
        mca.retrieve_file_over_http.assert_called_once_with(
            'http://download.alleninstitute.org/informatics-archive/annotation/ccf_2016/mouse_ccf/average_template/annotation_10.nrrd',
            'volumetric.nrrd')
    assert not Manifest.safe_mkdir.called, 'safe_mkdir should not have been called.'
    nrrd_read.assert_called_once_with('volumetric.nrrd')


@pytest.mark.parametrize("file_exists",
                         (True, False))
def test_file_download_server(mock_imports,
                              mca,
                              cache,
                             file_exists):
    nrrd_read, MCA = mock_imports

    @cacheable(reader=nrrd_read,
               pathfinder=Cache.pathfinder(file_name_position=3,
                                           secondary_file_name_position=1))
    def download_volumetric_data(data_path,
                                 file_name,
                                 voxel_resolution=None,
                                 save_file_path=None,
                                 release=None,
                                 coordinate_framework=None):
        url = mca.build_volumetric_data_download_url(data_path,
                                                     file_name,
                                                     voxel_resolution,
                                                     release,
                                                     coordinate_framework)

        mca.retrieve_file_over_http(url, save_file_path)

    with patch('os.path.exists',
               Mock(name="os.path.exists",
                    return_value=file_exists)) as mkdir:
        nrrd_read.reset_mock()
        
        download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                 'annotation_10.nrrd',
                                 MCA.VOXEL_RESOLUTION_10_MICRONS,
                                 'volumetric.nrrd',
                                 MCA.CCF_2016,
                                 strategy='create')

    mca.retrieve_file_over_http.assert_called_once_with(
        'http://download.alleninstitute.org/informatics-archive/annotation/ccf_2016/mouse_ccf/average_template/annotation_10.nrrd',
        'volumetric.nrrd')
    assert not Manifest.safe_mkdir.called, 'safe_mkdir should not have been called.'
    nrrd_read.assert_called_once_with('volumetric.nrrd')


@pytest.mark.parametrize("file_exists",
                         (True, False))
def test_file_download_cached_file(mock_imports,
                                   mca,
                                   cache,
                                   file_exists):
    nrrd_read, MCA = mock_imports

    @cacheable(reader=nrrd_read,
               pathfinder=Cache.pathfinder(file_name_position=3,
                                           secondary_file_name_position=1))
    def download_volumetric_data(data_path,
                                 file_name,
                                 voxel_resolution=None,
                                 save_file_path=None,
                                 release=None,
                                 coordinate_framework=None):
        url = mca.build_volumetric_data_download_url(data_path,
                                                     file_name,
                                                     voxel_resolution,
                                                     release,
                                                     coordinate_framework)

        mca.retrieve_file_over_http(url, save_file_path)

    with patch('os.path.exists',
               Mock(name="os.path.exists",
                    return_value=file_exists)) as mkdir:
        nrrd_read.reset_mock()

        download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                 'annotation_10.nrrd',
                                 MCA.VOXEL_RESOLUTION_10_MICRONS,
                                 'volumetric.nrrd',
                                 MCA.CCF_2016,
                                 strategy='file')

    assert not mca.retrieve_file_over_http.called, 'server should not have been called'
    assert not Manifest.safe_mkdir.called, 'safe_mkdir should not have been called.'
    nrrd_read.assert_called_once_with('volumetric.nrrd')


@pytest.mark.parametrize("file_exists",
                         (True, False))
def test_file_kwarg(mock_imports,
                    mca,
                    cache,
                    file_exists):
    nrrd_read, MCA = mock_imports

    @cacheable(reader=nrrd_read,
               pathfinder=Cache.pathfinder(file_name_position=3,
                                           secondary_file_name_position=1,
                                           path_keyword='save_file_path'))
    def download_volumetric_data(data_path,
                                 file_name,
                                 voxel_resolution=None,
                                 save_file_path=None,
                                 release=None,
                                 coordinate_framework=None):
        url = mca.build_volumetric_data_download_url(data_path,
                                                     file_name,
                                                     voxel_resolution,
                                                     release,
                                                     coordinate_framework)

        mca.retrieve_file_over_http(url, save_file_path)

    with patch('os.path.exists',
               Mock(name="os.path.exists",
                    return_value=file_exists)) as mkdir:
        nrrd_read.reset_mock()

        download_volumetric_data(MCA.AVERAGE_TEMPLATE,
                                 'annotation_10.nrrd',
                                 MCA.VOXEL_RESOLUTION_10_MICRONS,
                                 'volumetric.nrrd',
                                 MCA.CCF_2016,
                                 strategy='file',
                                 save_file_path='file.nrrd' )

    assert not mca.retrieve_file_over_http.called, 'server should not have been called'
    assert not Manifest.safe_mkdir.called, 'safe_mkdir should not have been called.'
    nrrd_read.assert_called_once_with('file.nrrd')


@pytest.mark.run('last')
def test_cleanup():
    import allensdk.api.queries.mouse_connectivity_api
    reload(allensdk.api.queries.mouse_connectivity_api)
    import allensdk.core.mouse_connectivity_cache
    reload(allensdk.core.mouse_connectivity_cache)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import scipy.stats as st
import numpy as np
import pandas as pd
from .stimulus_analysis import StimulusAnalysis
import logging
import h5py
from . import observatory_plots as oplots
from . import circle_plots as cplots
from .brain_observatory_exceptions import MissingStimulusException

class NaturalScenes(StimulusAnalysis):
    """ Perform tuning analysis specific to natural scenes stimulus.

    Parameters
    ----------
    data_set: BrainObservatoryNwbDataSet object
    """

    _log = logging.getLogger('allensdk.brain_observatory.natural_scenes')

    def __init__(self, data_set, **kwargs):
        super(NaturalScenes, self).__init__(data_set, **kwargs)

        self._number_scenes = StimulusAnalysis._PRELOAD
        self._sweeplength = StimulusAnalysis._PRELOAD
        self._interlength = StimulusAnalysis._PRELOAD
        self._extralength = StimulusAnalysis._PRELOAD

    @property
    def number_scenes(self):
        if self._number_scenes is StimulusAnalysis._PRELOAD:
            self.populate_stimulus_table()

        return self._number_scenes

    @property
    def sweeplength(self):
        if self._sweeplength is StimulusAnalysis._PRELOAD:
            self.populate_stimulus_table()

        return self._sweeplength

    @property
    def interlength(self):
        if self._interlength is StimulusAnalysis._PRELOAD:
            self.populate_stimulus_table()

        return self._interlength

    @property
    def extralength(self):
        if self._extralength is StimulusAnalysis._PRELOAD:
            self.populate_stimulus_table()

        return self._extralength

    def populate_stimulus_table(self):
        self._stim_table = self.data_set.get_stimulus_table('natural_scenes')
        self._number_scenes = len(np.unique(self._stim_table.frame))
        self._sweeplength = self._stim_table.end.iloc[
            1] - self._stim_table.start.iloc[1]
        self._interlength = 4 * self._sweeplength
        self._extralength = self._sweeplength

    def get_response(self):
        ''' Computes the mean response for each cell to each stimulus condition.  Return is
        a (# scenes, # cells, 3) np.ndarray.  The final dimension
        contains the mean response to the condition (index 0), standard error of the mean of the response
        to the condition (index 1), and the number of trials with a significant (p < 0.05) response 
        to that condition (index 2).

        Returns
        -------
        Numpy array storing mean responses.
        '''
        NaturalScenes._log.info("Calculating mean responses")

        response = np.empty((self.number_scenes, self.numbercells + 1, 3))

        def ptest(x):
            return len(np.where(x < (0.05 / (self.number_scenes - 1)))[0])

        for ns in range(self.number_scenes):
            subset_response = self.mean_sweep_response[
                self.stim_table.frame == (ns - 1)]
            subset_pval = self.pval[self.stim_table.frame == (ns - 1)]
            response[ns, :, 0] = subset_response.mean(axis=0)
            response[ns, :, 1] = subset_response.std(
                axis=0) / np.sqrt(len(subset_response))
            response[ns, :, 2] = subset_pval.apply(ptest, axis=0)

        return response

    def get_peak(self):
        ''' Computes metrics about peak response condition for each cell.

        Returns
        -------
        Pandas data frame with the following fields ('_ns' suffix is for
        natural scene):
            * scene_ns (scene number)
            * reliability_ns
            * peak_dff_ns (peak dF/F)
            * ptest_ns
            * p_run_ns
            * run_modulation_ns
            * time_to_peak_ns
        '''
        NaturalScenes._log.info('Calculating peak response properties')
        peak = pd.DataFrame(index=range(self.numbercells), columns=('scene_ns', 'reliability_ns', 'peak_dff_ns',
                                                                    'ptest_ns', 'p_run_ns', 'run_modulation_ns', 
                                                                    'time_to_peak_ns', 
                                                                    'cell_specimen_id','image_selectivity_ns'))
        cids = self.data_set.get_cell_specimen_ids()

        for nc in range(self.numbercells):
            nsp = np.argmax(self.response[1:, nc, 0])
            peak.cell_specimen_id.iloc[nc] = cids[nc]
            peak.scene_ns[nc] = nsp
#            peak.response_reliability_ns[nc] = self.response[
#                nsp + 1, nc, 2] / 0.50  # assume 50 trials
            peak.peak_dff_ns[nc] = self.response[nsp + 1, nc, 0]
#            subset = self.mean_sweep_response[self.stim_table.frame == nsp]
#            subset_stat = subset[subset.dx < 2]
#            subset_run = subset[subset.dx >= 2]
#            if (len(subset_run) > 5) & (len(subset_stat) > 5):
#                (_, peak.p_run_ns[nc]) = st.ks_2samp(
#                    subset_run[str(nc)], subset_stat[str(nc)])
#                peak.run_modulation_ns[nc] = subset_run[
#                    str(nc)].mean() / subset_stat[str(nc)].mean()
#            else:
#                peak.p_run_ns[nc] = np.NaN
#                peak.run_modulation_ns[nc] = np.NaN
            groups = []
            for im in range(self.number_scenes):
                subset = self.mean_sweep_response[
                    self.stim_table.frame == (im - 1)]
                groups.append(subset[str(nc)].values)
            (_, peak.ptest_ns[nc]) = st.f_oneway(*groups)
            test = self.sweep_response[
                self.stim_table.frame == nsp][str(nc)].mean()
            peak.time_to_peak_ns[nc] = (
                np.argmax(test) - self.interlength) / self.acquisition_rate
            
            #running modulation
            subset = self.mean_sweep_response[self.stim_table.frame==nsp]            
            subset_run = subset[subset.dx>=1]
            subset_stat = subset[subset.dx<1]
            if (len(subset_run)>4) & (len(subset_stat)>4):
                (_,peak.p_run_ns.iloc[nc]) = st.ttest_ind(subset_run[str(nc)], subset_stat[str(nc)], equal_var=False)
                
                if subset_run[str(nc)].mean()>subset_stat[str(nc)].mean():
                    peak.run_modulation_ns.iloc[nc] = (subset_run[str(nc)].mean() - subset_stat[str(nc)].mean())/np.abs(subset_run[str(nc)].mean())
                elif subset_run[str(nc)].mean()<subset_stat[str(nc)].mean():
                    peak.run_modulation_ns.iloc[nc] = -1*((subset_stat[str(nc)].mean() - subset_run[str(nc)].mean())/np.abs(subset_stat[str(nc)].mean()))
            else:
                peak.p_run_ns.iloc[nc] = np.NaN
                peak.run_modulation_ns.iloc[nc] = np.NaN                
            
            #reliability
            subset = self.sweep_response[self.stim_table.frame==nsp]            
            corr_matrix = np.empty((len(subset),len(subset)))
            for i in range(len(subset)):
                for j in range(len(subset)):
                    r,p = st.pearsonr(subset[str(nc)].iloc[i][28:42], subset[str(nc)].iloc[j][28:42])
                    corr_matrix[i,j] = r
            mask = np.ones((len(subset), len(subset)))
            for i in range(len(subset)):
                for j in range(len(subset)):
                    if i>=j:
                        mask[i,j] = np.NaN
            corr_matrix *= mask
            peak.reliability_ns.iloc[nc] = np.nanmean(corr_matrix)
            
            #image selectivity
            fmin = self.response[1:,nc,0].min()
            fmax = self.response[1:,nc,0].max()
            rtj = np.empty((1000,1))
            for j in range(1000):
                thresh = fmin + j*((fmax-fmin)/1000.)
                theta = np.empty((118,1))
                for im in range(118):
                    if self.response[im+1,nc,0] > thresh:  #im+1 to only look at images, not blanksweep
                        theta[im] = 1
                    else:
                        theta[im] = 0
                rtj[j] = theta.mean()
            
            biga = rtj.mean()
            bigs = 1 - (2*biga)
            peak.image_selectivity_ns.iloc[nc] = bigs

        return peak

    def plot_time_to_peak(self, 
                          p_value_max=oplots.P_VALUE_MAX, 
                          color_map=oplots.STIMULUS_COLOR_MAP):
        stimulus_table = self.data_set.get_stimulus_table('natural_scenes')

        resps = []

        for index, row in self.peak.iterrows():    
            mean_response = self.sweep_response.ix[stimulus_table.frame==row.scene_ns][str(index)].mean()
            resps.append((mean_response - mean_response.mean() / mean_response.std()))

        mean_responses = np.array(resps)

        sorted_table = self.peak[self.peak.ptest_ns < p_value_max].sort(columns='time_to_peak_ns')
        cell_order = sorted_table.index

        # time to peak is relative to stimulus start in seconds
        ttps = sorted_table.time_to_peak_ns.values + self.interlength / self.acquisition_rate
        msrs_sorted = mean_responses[cell_order,:]

        oplots.plot_time_to_peak(msrs_sorted, ttps,
                                 0, (2*self.interlength + self.sweeplength) / self.acquisition_rate,
                                 (self.interlength) / self.acquisition_rate, 
                                 (self.interlength + self.sweeplength) / self.acquisition_rate, 
                                 color_map)

    def open_corona_plot(self, cell_specimen_id=None, cell_index=None):
        cell_index = self.row_from_cell_id(cell_specimen_id, cell_index)

        df = self.mean_sweep_response[str(cell_index)]
        data = df.values

        st = self.data_set.get_stimulus_table('natural_scenes')
        mask = st[st.frame >= 0].index

        cmin = self.response[0,cell_index,0]
        cmax = data.mean() + data.std()*3

        cp = cplots.CoronaPlotter()
        cp.plot(st.frame.ix[mask].values, 
                data=df.ix[mask].values,
                clim=[cmin, cmax])
        cp.show_arrow()

    def reshape_response_array(self):
        '''
        :return: response array in cells x stim x repetition for noise correlations
        '''

        mean_sweep_response = self.mean_sweep_response.values[:, :self.numbercells]

        stim_table = self.stim_table
        frames = np.unique(stim_table.frame.values)

        reps = [len(np.where(stim_table.frame.values == frame)[0]) for frame in frames]
        Nreps = min(reps) # just in case there are different numbers of repetitions

        response_new = np.zeros((self.numbercells, self.number_scenes), dtype='object')
        for i, frame in enumerate(frames):
            ind = np.where(stim_table.frame.values == frame)[0][:Nreps]
            for c in range(self.numbercells):
                response_new[c, i] = mean_sweep_response[ind, c]

        return response_new

    def get_signal_correlation(self, corr='spearman'):
        logging.debug("Calculating signal correlations")

        response = self.response[:, :, 0].T
        response = response[:self.numbercells, :]
        N, Nstim = response.shape

        signal_corr = np.zeros((N, N))
        signal_p = np.empty((N, N))
        if corr == 'pearson':
            for i in range(N):
                for j in range(i, N): # matrix is symmetric
                    signal_corr[i, j], signal_p[i, j] = st.pearsonr(response[i], response[j])

        elif corr == 'spearman':
            for i in range(N):
                for j in range(i, N): # matrix is symmetric
                    signal_corr[i, j], signal_p[i, j] = st.spearmanr(response[i], response[j])

        else:
            raise Exception('correlation should be pearson or spearman')

        signal_corr = np.triu(signal_corr) + np.triu(signal_corr, 1).T  # fill in lower triangle
        signal_p = np.triu(signal_p) + np.triu(signal_p, 1).T  # fill in lower triangle

        return signal_corr, signal_p

    def get_representational_similarity(self, corr='spearman'):
        logging.debug("Calculating representational similarity")

        response = self.response[:, :, 0]
        response = response[:, :self.numbercells]
        Nstim, N = response.shape

        rep_sim = np.zeros((Nstim, Nstim))
        rep_sim_p = np.empty((Nstim, Nstim))
        if corr == 'pearson':
            for i in range(Nstim):
                for j in range(i, Nstim): # matrix is symmetric
                    rep_sim[i, j], rep_sim_p[i, j] = st.pearsonr(response[i], response[j])

        elif corr == 'spearman':
            for i in range(Nstim):
                for j in range(i, Nstim): # matrix is symmetric
                    rep_sim[i, j], rep_sim_p[i, j] = st.spearmanr(response[i], response[j])

        else:
            raise Exception('correlation should be pearson or spearman')

        rep_sim = np.triu(rep_sim) + np.triu(rep_sim, 1).T # fill in lower triangle
        rep_sim_p = np.triu(rep_sim_p) + np.triu(rep_sim_p, 1).T  # fill in lower triangle

        return rep_sim, rep_sim_p

    def get_noise_correlation(self, corr='spearman'):
        logging.debug("Calculating noise correlations")

        response = self.reshape_response_array()
        noise_corr = np.zeros((self.numbercells, self.numbercells, self.number_scenes))
        noise_corr_p = np.zeros((self.numbercells, self.numbercells, self.number_scenes))

        if corr == 'pearson':
            for k in range(self.number_scenes):
                for i in range(self.numbercells):
                    for j in range(i, self.numbercells):
                        noise_corr[i, j, k], noise_corr_p[i, j, k] = st.pearsonr(response[i, k], response[j, k])

                noise_corr[:, :, k] = np.triu(noise_corr[:, :, k]) + np.triu(noise_corr[:, :, k], 1).T
                noise_corr_p[:, :, k] = np.triu(noise_corr_p[:, :, k]) + np.triu(noise_corr_p[:, :, k], 1).T

        elif corr == 'spearman':
            for k in range(self.number_scenes):
                for i in range(self.numbercells):
                    for j in range(i, self.numbercells):
                        noise_corr[i, j, k], noise_corr_p[i, j, k] = st.spearmanr(response[i, k], response[j, k])

                noise_corr[:, :, k] = np.triu(noise_corr[:, :, k]) + np.triu(noise_corr[:, :, k], 1).T
                noise_corr_p[:, :, k] = np.triu(noise_corr_p[:, :, k]) + np.triu(noise_corr_p[:, :, k], 1).T

        else:
            raise Exception('correlation should be pearson or spearman')

        return noise_corr, noise_corr_p

    @staticmethod
    def from_analysis_file(data_set, analysis_file):
        ns = NaturalScenes(data_set)
        ns.populate_stimulus_table()

        try:
            ns._sweep_response = pd.read_hdf(analysis_file, "analysis/sweep_response_ns")
            ns._mean_sweep_response = pd.read_hdf(analysis_file, "analysis/mean_sweep_response_ns")
            ns._peak = pd.read_hdf(analysis_file, "analysis/peak")

            with h5py.File(analysis_file, "r") as f:
                ns._response = f["analysis/response_ns"].value
                ns._binned_dx_sp = f["analysis/binned_dx_sp"].value
                ns._binned_cells_sp = f["analysis/binned_cells_sp"].value
                ns._binned_dx_vis = f["analysis/binned_dx_vis"].value
                ns._binned_cells_vis = f["analysis/binned_cells_vis"].value

                if "analysis/noise_corr_ns" in f:
                    ns.noise_correlation = f["analysis/noise_corr_ns"].value
                if "analysis/signal_corr_ns" in f:
                    ns.signal_correlation = f["analysis/signal_corr_ns"].value
                if "analysis/rep_similarity_ns" in f:
                    ns.representational_similarity = f["analysis/rep_similarity_ns"].value

        except Exception as e:
            raise MissingStimulusException(e.args)

        return ns




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest

import os
from allensdk.test_utilities.temp_dir import fn_temp_dir

@pytest.mark.skipif(True,
                    reason="partial testing")
def test_notebook(fn_temp_dir):


    # coding: utf-8

    # ## Mouse Connectivity
    #
    # This notebook demonstrates how to access and manipulate data in the Allen Mouse Brain Connectivity Atlas. The `MouseConnectivityCache` AllenSDK class provides methods for downloading metadata about experiments, including their viral injection site and the mouse's transgenic line. You can request information either as a Pandas DataFrame or a simple list of dictionaries.
    #
    # An important feature of the `MouseConnectivityCache` is how it stores and retrieves data for you. By default, it will create (or read) a manifest file that keeps track of where various connectivity atlas data are stored. If you request something that has not already been downloaded, it will download it and store it in a well known location.
    #
    # Download this notebook in .ipynb format <a href='mouse_connectivity.ipynb'>here</a>.

    # In[1]:

    from allensdk.core.mouse_connectivity_cache import MouseConnectivityCache

    # The manifest file is a simple JSON file that keeps track of all of
    # the data that has already been downloaded onto the hard drives.
    # If you supply a relative path, it is assumed to be relative to your
    # current working directory.
    mcc = MouseConnectivityCache(manifest_file=os.path.join(fn_temp_dir, 'connectivity/mouse_connectivity_manifest.json'))

    # open up a list of all of the experiments
    all_experiments = mcc.get_experiments(dataframe=True)
    print "%d total experiments" % len(all_experiments)

    # take a look at what we know about an experiment with a primary motor injection
    all_experiments.loc[122642490]


    # `MouseConnectivityCache` has a method for retrieving the adult mouse structure tree as an `StructureTree` class instance. This is a wrapper around a list of dictionaries, where each dictionary describes a structure. It is principally useful for looking up structures by their properties.

    # In[2]:

    # pandas for nice tables
    import pandas as pd

    # grab the StructureTree instance
    structure_tree = mcc.get_structure_tree()

    # get info on some structures
    structures = structure_tree.get_structures_by_name(['Primary visual area', 'Hypothalamus'])
    pd.DataFrame(structures)


    # As a convenience, structures are grouped in to named collections called "structure sets". These sets can be used to quickly gather a useful subset of structures from the tree. The criteria used to define structure sets are eclectic; a structure set might list:
    #
    # * structures that were used in a particular project.
    # * structures that coarsely partition the brain.
    # * structures that bear functional similarity.
    #
    # or something else entirely. To view all of the available structure sets along with their descriptions, follow this [link](http://api.brain-map.org/api/v2/data/StructureSet/query.json). To see only structure sets relevant to the adult mouse brain, use the StructureTree:

    # In[3]:

    from allensdk.api.queries.ontologies_api import OntologiesApi

    oapi = OntologiesApi()

    # get the ids of all the structure sets in the tree
    structure_set_ids = structure_tree.get_structure_sets()

    # query the API for information on those structure sets
    pd.DataFrame(oapi.get_structure_sets(structure_set_ids))


    # On the connectivity atlas web site, you'll see that we show most of our data at a fairly coarse structure level. We did this by creating a structure set of ~300 structures, which we call the "summary structures". We can use the structure tree to get all of the structures in this set:

    # In[4]:

    # From the above table, "Mouse Connectivity - Summary" has id 167587189
    summary_structures = structure_tree.get_structures_by_set_id([167587189])
    pd.DataFrame(summary_structures)


    # This is how you can filter experiments by transgenic line:

    # In[5]:

    # fetch the experiments that have injections in the isocortex of cre-positive mice
    isocortex = structure_tree.get_structures_by_name(['Isocortex'])[0]
    cre_cortical_experiments = mcc.get_experiments(cre=True, 
                                                    injection_structure_ids=[isocortex['id']])

    print "%d cre cortical experiments" % len(cre_cortical_experiments)

    # same as before, but restrict the cre line
    rbp4_cortical_experiments = mcc.get_experiments(cre=[ 'Rbp4-Cre_KL100' ], 
                                                    injection_structure_ids=[isocortex['id']])


    print "%d Rbp4 cortical experiments" % len(rbp4_cortical_experiments)


    # ## Structure Signal Unionization
    #
    # The ProjectionStructureUnionizes API data tells you how much signal there was in a given structure and experiment. It contains the density of projecting signal, volume of projecting signal, and other information. `MouseConnectivityCache` provides methods for querying and storing this data.

    # In[6]:

    # find wild-type injections into primary visual area
    visp = structure_tree.get_structures_by_acronym(['VISp'])[0]
    visp_experiments = mcc.get_experiments(cre=False, 
                                           injection_structure_ids=[visp['id']])

    print "%d VISp experiments" % len(visp_experiments)

    structure_unionizes = mcc.get_structure_unionizes([ e['id'] for e in visp_experiments ], 
                                                      is_injection=False,
                                                      structure_ids=[isocortex['id']],
                                                      include_descendants=True)

    print "%d VISp non-injection, cortical structure unionizes" % len(structure_unionizes)


    # In[7]:

    structure_unionizes.head()


    # This is a rather large table, even for a relatively small number of experiments.  You can filter it down to a smaller list of structures like this.

    # In[8]:

    dense_unionizes = structure_unionizes[ structure_unionizes.projection_density > .5 ]
    large_unionizes = dense_unionizes[ dense_unionizes.volume > .5 ]
    large_structures = pd.DataFrame(structure_tree.node(large_unionizes.structure_id))

    print "%d large, dense, cortical, non-injection unionizes, %d structures" % ( len(large_unionizes), len(large_structures) )

    print large_structures.name

    large_unionizes


    # ## Generating a Projection Matrix
    # The `MouseConnectivityCache` class provides a helper method for converting ProjectionStructureUnionize records for a set of experiments and structures into a matrix.  This code snippet demonstrates how to make a matrix of projection density values in auditory sub-structures for cre-negative VISp experiments. 

    # In[9]:

    import numpy as np
    import warnings
    warnings.filterwarnings('ignore')
    
    visp_experiment_ids = [ e['id'] for e in visp_experiments ]
    ctx_children = structure_tree.child_ids( [isocortex['id']] )[0]

    pm = mcc.get_projection_matrix(experiment_ids = visp_experiment_ids, 
                                   projection_structure_ids = ctx_children,
                                   hemisphere_ids= [2], # right hemisphere, ipsilateral
                                   parameter = 'projection_density')

    row_labels = pm['rows'] # these are just experiment ids
    column_labels = [ c['label'] for c in pm['columns'] ] 
    matrix = pm['matrix']



    # ## Manipulating Grid Data
    #
    # The `MouseConnectivityCache` class also helps you download and open every experiment's projection grid data volume. By default it will download 25um volumes, but uou could also download data at other resolutions if you prefer (10um, 50um, 100um).
    #
    # This demonstrates how you can load the projection density for a particular experiment. It also shows how to download the template volume to which all grid data is registered. Voxels in that template have been structurally annotated by neuroanatomists and stored in a separate annotation volume image.

    # In[10]:

    experiment_id = 181599674

    # projection density: number of projecting pixels / voxel volume
    pd, pd_info = mcc.get_projection_density(experiment_id)

    # injection density: number of projecting pixels in injection site / voxel volume
    ind, ind_info = mcc.get_injection_density(experiment_id)

    # injection fraction: number of pixels in injection site / voxel volume
    inf, inf_info = mcc.get_injection_fraction(experiment_id)

    # data mask:
    # binary mask indicating which voxels contain valid data
    dm, dm_info = mcc.get_data_mask(experiment_id)

    template, template_info = mcc.get_template_volume()
    annot, annot_info = mcc.get_annotation_volume()

    print(pd_info)
    print(pd.shape, template.shape, annot.shape)




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
import pytest
from mock import patch, MagicMock, call
from collections import Counter

def make_bo_api():
    from allensdk.api.queries.brain_observatory_api import BrainObservatoryApi

    endpoint = os.environ['TEST_API_ENDPOINT'] if 'TEST_API_ENDPOINT' in os.environ else 'http://twarehouse-backup'
    return BrainObservatoryApi(endpoint)

@pytest.fixture
def bo_api():
    bo = make_bo_api()
    bo.json_msg_query = MagicMock(name='json_msg_query')

    return bo


@pytest.fixture
def bo_api5():
    rows_per_message = 2000
    num_messages = 5
    _msg = [{'whatever': True}] * rows_per_message
    import allensdk.core.json_utilities as ju
    ju.read_url_get = MagicMock(name='json_msg_query',
                                side_effect = [{'msg': _msg}] * num_messages)

    return { 'read_url_get': ju.read_url_get,
             'bo': make_bo_api() }


@pytest.fixture
def bo_api_save_ophys():
    bo = make_bo_api()
    bo.json_msg_query = \
        MagicMock(name='json_msg_query',
                  return_value=[{'download_link': '/url/path/to/file'}])

    return bo


@pytest.fixture
def mock_containers():
    containers = [
        {'targeted_structure': {'acronym': 'CBS'},
         'imaging_depth': 100,
         'specimen': {
            'donor': {'transgenic_lines': [{'name': 'Shiny'}]}}
         },
        {'targeted_structure': {'acronym': 'NBC'},
         'imaging_depth': 200,
         'specimen': {
            'donor': {'transgenic_lines': [{'name': 'Don'}]}}
         }
    ]

    return containers


@pytest.fixture
def mock_ophys_experiments():
    experiments = [
        {'experiment_container_id': 1,
         'targeted_structure': {'acronym': 'CBS'},
         'imaging_depth': 100,
         'specimen': {'donor': {
             'transgenic_lines': [{'name': 'Shiny'}]}},
         'stimulus_name': 'three_session_B',
         },
        {'experiment_container_id': 2,
         'targeted_structure': {'acronym': 'NBC'},
         'imaging_depth': 200,
         'specimen': {'donor': {
             'transgenic_lines': [{'name': 'Don'}]}},
         'stimulus_name': 'three_session_C',
         'experiment_container': { 'failed': False }
         },
        {'experiment_container_id': 2,
         'targeted_structure': {'acronym': 'NBC'},
         'imaging_depth': 200,
         'specimen': {'donor': {
             'transgenic_lines': [{'name': 'Don'}]}},
         'stimulus_name': 'three_session_C',
         'experiment_container': { 'failed': True }
         }
    ]

    return experiments


@pytest.fixture
def mock_specimens():
    specimens = [
        {"experiment_container_id": 511498500,
         "cell_specimen_id": 517394843
         },
        {"experiment_container_id": 511498742,
         "cell_specimen_id": 517398740,
         "failed_experiment_container": False
         },
        {"experiment_container_id": 511498501,
         "cell_specimen_id": 517394874,
         "failed_experiment_container": True
         }
    ]
    return specimens


def test_list_isi_experiments(bo_api):
    bo_api.list_isi_experiments()
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::IsiExperiment,rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_isi_experiments(bo_api):
    isi_experiment_id = 503316697
    bo_api.get_isi_experiments(isi_experiment_id)
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::IsiExperiment,rma::criteria,[id$in503316697],"
        "rma::include,"
        "experiment_container(ophys_experiments,targeted_structure),"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_ophys_experiments_one_id(bo_api):
    ophys_experiment_id = 502066273
    bo_api.get_ophys_experiments(ophys_experiment_id)
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::OphysExperiment,rma::criteria,[id$in502066273],"
        "rma::include,experiment_container,"
        "well_known_files(well_known_file_type),targeted_structure,"
        "specimen(donor(age,transgenic_lines)),"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_experiment_container_metrics(bo_api):
    tid = 511510627
    bo_api.get_experiment_container_metrics(tid)
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiCamExperimentContainerMetric,"
        "rma::criteria,[id$in511510627],"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_experiment_containers(bo_api):
    tid = 511510753
    bo_api.get_experiment_containers(tid)
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ExperimentContainer,rma::criteria,[id$in511510753],"
        "rma::include,ophys_experiments,isi_experiment,"
        "specimen(donor(conditions,age,transgenic_lines)),targeted_structure,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_column_definitions(bo_api):
    api_class_name = bo_api.quote_string('ApiTbiDonorMetric')
    bo_api.get_column_definitions(api_class_name=api_class_name)
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiColumnDefinition,"
        "rma::criteria,[api_class_name$eq'ApiTbiDonorMetric'],"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_list_column_definition_class_names(bo_api):
    bo_api.list_column_definition_class_names()
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiColumnDefinition,"
        "rma::options"
        "[only$eq'api_class_name'][num_rows$eq'all'][count$eqfalse]")


def test_get_stimulus_mappings_no_ids(bo_api):
    bo_api.get_stimulus_mappings()
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiCamStimulusMapping,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_stimulus_mappings_one_id(bo_api):
    ids = 15
    bo_api.get_stimulus_mappings(stimulus_mapping_ids=ids)
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiCamStimulusMapping,"
        "rma::criteria,[id$in15],"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_stimulus_mappings_two_ids(bo_api):
    ids = [15, 43]
    bo_api.get_stimulus_mappings(stimulus_mapping_ids=ids)
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiCamStimulusMapping,"
        "rma::criteria,[id$in15,43],"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_cell_metrics_no_ids(bo_api):
    list(bo_api.get_cell_metrics())
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiCamCellMetric,"
        "rma::options[num_rows$eq2000][start_row$eq0][count$eqfalse]")


def test_get_cell_metrics_one_ids(bo_api):
    tid = 517394843
    list(bo_api.get_cell_metrics(cell_specimen_ids=tid))
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiCamCellMetric,"
        "rma::criteria,[cell_specimen_id$in517394843],"
        "rma::options[num_rows$eq2000][start_row$eq0][count$eqfalse]")


def test_get_cell_metrics_two_ids(bo_api):
    ids = [517394843, 517394850]
    res = list(bo_api.get_cell_metrics(cell_specimen_ids=ids))
    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::ApiCamCellMetric,"
        "rma::criteria,[cell_specimen_id$in517394843,517394850],"
        "rma::options[num_rows$eq2000][start_row$eq0][count$eqfalse]")


def test_get_cell_metrics_five_messages(bo_api5):
    ids = [517394843, 517394850]
    list(bo_api5['bo'].get_cell_metrics(cell_specimen_ids=ids))

    base_query = \
       (bo_api5['bo'].api_url + '/api/v2/data/query.json?q='
        'model::ApiCamCellMetric,'
        'rma::criteria,%5Bcell_specimen_id$in517394843,517394850%5D,'
        'rma::options%5Bnum_rows$eq2000%5D%5Bstart_row$eq{}%5D%5Bcount$eqfalse%5D')
    expected_calls = map(lambda c: call(base_query.format(c)),
                         [0, 2000, 4000, 6000, 8000, 10000])

    assert bo_api5['read_url_get'].call_args_list == expected_calls


def test_filter_experiment_containers_no_filters(bo_api, mock_containers):
    containers = bo_api.filter_experiment_containers(mock_containers)
    assert len(containers) == 2


def test_filter_experiment_containers_depth_filter(bo_api, mock_containers):
    containers = bo_api.filter_experiment_containers(mock_containers,
                                                     imaging_depths=[100])
    assert len(containers) == 1


def test_filter_experiment_containers_structures_filter(bo_api, mock_containers):
    containers = \
        bo_api.filter_experiment_containers(
            mock_containers,
            targeted_structures=['CBS'])
    assert len(containers) == 1


def test_filter_experiment_containers_lines_all_filters(bo_api, mock_containers):
    containers = \
        bo_api.filter_experiment_containers(mock_containers,
                                            imaging_depths=[200],
                                            targeted_structures=['NBC'],
                                            transgenic_lines=['Don'])
    assert len(containers) == 1


def test_filter_ophys_experiments_no_filters(bo_api, mock_ophys_experiments):
    experiments = bo_api.filter_ophys_experiments(mock_ophys_experiments)
    assert len(experiments) == 2


def test_filter_ophys_experiments_container_id(bo_api, mock_ophys_experiments):
    experiments = bo_api.filter_ophys_experiments(mock_ophys_experiments,
                                                  experiment_container_ids=[1])
    assert len(experiments) == 1


def test_filter_ophys_experiments_stimuli(bo_api, mock_ophys_experiments):
    experiments = bo_api.filter_ophys_experiments(mock_ophys_experiments,
                                                  stimuli=['static_gratings'])
    assert len(experiments) == 1


def test_filter_cell_specimens(bo_api, mock_specimens):
    specimens = bo_api.filter_cell_specimens(mock_specimens, include_failed=True)
    assert specimens == mock_specimens

    specimens = bo_api.filter_cell_specimens(mock_specimens)
    assert len(specimens) == 2

    specimens = bo_api.filter_cell_specimens(
        mock_specimens, ids=[mock_specimens[0]['cell_specimen_id']])
    assert len(specimens) == 1
    assert specimens[0] == mock_specimens[0]

    cnt = Counter()
    for sp in mock_specimens:
        cnt[sp['experiment_container_id']] += 1

    ecid = mock_specimens[0]['experiment_container_id']
    specimens = bo_api.filter_cell_specimens(
        mock_specimens, experiment_container_ids=[ecid])
    assert len(specimens) == cnt[ecid]
    assert specimens[0] == mock_specimens[0]


def test_save_ophys_experiment_data(bo_api_save_ophys):
    bo_api = bo_api_save_ophys

    with patch('allensdk.config.manifest.Manifest.safe_mkdir') as mkdir:
        bo_api.retrieve_file_over_http = \
            MagicMock(name='retrieve_file_over_http')
        bo_api.save_ophys_experiment_data(1, '/path/to/filename')

        mkdir.assert_called_once_with('/path/to')

    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::WellKnownFile,"
        "rma::criteria,"
        "[attachable_id$eq1],well_known_file_type[name$eqNWBOphys],"
        "rma::options[num_rows$eq'all'][count$eqfalse]")
    bo_api.retrieve_file_over_http.assert_called_with(
        bo_api.api_url +  '/url/path/to/file',
        '/path/to/filename')

def test_get_cell_specimen_id_mapping(bo_api_save_ophys):
    bo_api = bo_api_save_ophys

    with patch('pandas.read_csv') as readcsv:
        bo_api.retrieve_file_over_http = \
            MagicMock(name='retrieve_file_over_http')
        bo_api.get_cell_specimen_id_mapping('/path/to/filename', 1)

        readcsv.assert_called_once_with('/path/to/filename')

    bo_api.json_msg_query.assert_called_once_with(
        bo_api.api_url + "/api/v2/data/query.json?q="
        "model::WellKnownFile,"
        "rma::criteria,"
        "[id$eq1],well_known_file_type[name$eqOphysCellSpecimenIdMapping],"
        "rma::options[num_rows$eq'all'][count$eqfalse]")
    bo_api.retrieve_file_over_http.assert_called_with(
        bo_api.api_url + '/url/path/to/file',
        '/path/to/filename')
